#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.03 17:00:00                  #
# ================================================== #

from openai import OpenAI

from pygpt_net.core.types import (
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_IMAGE,
    MODE_VISION,
    MODE_RESEARCH,
    MODE_COMPUTER,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.types.chunk import ChunkType
from pygpt_net.item.model import ModelItem

from .audio import Audio
from .assistants import Assistants
from .chat import Chat
from .completion import Completion
from .computer import Computer
from .container import Container
from .image import Image
from .remote_tools import RemoteTools
from .responses import Responses
from .realtime import Realtime
from .store import Store
from .summarizer import Summarizer
from .tools import Tools
from .vision import Vision
from .video import Video


class ApiOpenAI:

    def __init__(self, window=None):
        """
        OpenAI API wrapper core

        :param window: Window instance
        """
        self.window = window
        self.assistants = Assistants(window)
        self.audio = Audio(window)
        self.chat = Chat(window)
        self.completion = Completion(window)
        self.container = Container(window)
        self.computer = Computer(window)
        self.image = Image(window)
        self.remote_tools = RemoteTools(window)
        self.responses = Responses(window)
        self.realtime = Realtime(window)
        self.store = Store(window)
        self.summarizer = Summarizer(window)
        self.tools = Tools(window)
        self.vision = Vision(window)
        self.video = Video(window)
        self.client = None
        self.locked = False
        self.last_client_args = None  # last client args used, for debug purposes

    def get_client(
            self,
            mode: str = MODE_CHAT,
            model: ModelItem = None
    ) -> OpenAI:
        """
        Return OpenAI client

        :param mode: Mode
        :param model: Model
        :return: OpenAI client
        """
        # prepare client args by mode and model provider
        args = self.window.core.models.prepare_client_args(mode, model)
        self.client = OpenAI(**args)
        self.last_client_args = args  # store last args for debug
        return self.client

    def call(
            self,
            context: BridgeContext,
            extra: dict = None,
            rt_signals=None
    ) -> bool:
        """
        Call OpenAI API

        :param context: Bridge context
        :param extra: Extra arguments
        :param rt_signals: Realtime signals for audio streaming
        :return: result
        """
        mode = context.mode
        parent_mode = context.parent_mode  # real mode (global)
        prompt = context.prompt
        stream = context.stream
        model = context.model  # model instance (item, not id)
        system_prompt = context.system_prompt
        assistant_id = context.assistant_id
        tools_outputs = context.tools_outputs
        max_tokens = context.max_tokens  # max output tokens
        is_expert_call = context.is_expert_call
        preset = context.preset  # PresetItem, used for expert calls

        ctx = context.ctx
        ai_name = ctx.output_name
        thread_id = ctx.thread  # from ctx
        ctx.chunk_type = ChunkType.API_CHAT  # default: ChatCompletion API

        # --- Responses API ----
        use_responses_api = self.responses.is_enabled(model, mode, parent_mode, is_expert_call, preset)
        if use_responses_api:
            ctx.chunk_type = ChunkType.API_CHAT_RESPONSES # Responses API

        fixtures = self.window.controller.debug.fixtures

        # get model id
        model_id = None
        if model is not None:
            model_id = model.id
            if max_tokens > model.tokens:  # check max output tokens
                max_tokens = model.tokens

        response = None
        used_tokens = 0
        context.max_tokens = max_tokens  # update max output tokens
        file_ids = context.file_ids  # uploaded files IDs (assistant mode only)

        # completion
        if mode == MODE_COMPLETION:
            ctx.chunk_type = ChunkType.API_COMPLETION
            response = self.completion.send(
                context=context,
                extra=extra,
            )
            used_tokens = self.completion.get_used_tokens()

        # chat, audio (OpenAI) | research (deep research, Perplexity)
        elif mode in [
            MODE_CHAT,
            MODE_AUDIO,
            MODE_RESEARCH,
            MODE_COMPUTER,
        ]:
            if mode == MODE_AUDIO and stream:
                # Realtime API for audio streaming
                is_realtime = self.realtime.begin(
                    context=context,
                    model=model,
                    extra=extra or {},
                    rt_signals=rt_signals
                )
                if is_realtime:
                    return True

            if fixtures.is_enabled("stream"):  # fake stream for testing
                use_responses_api = False
                response = fixtures.get_stream_generator(ctx)
            else:
                # responses API
                if use_responses_api:
                    response = self.responses.send(
                        context=context,
                        extra=extra,
                    )
                    used_tokens = self.responses.get_used_tokens()
                else:
                    # chat completion API
                    response = self.chat.send(
                        context=context,
                        extra=extra,
                    )
                    if hasattr(response, "citations"):
                        if response.citations:
                            ctx.urls = response.citations
                    used_tokens = self.chat.get_used_tokens()

            self.vision.append_images(ctx)  # append images to ctx if provided

        # image / video
        elif mode == MODE_IMAGE:
            media_mode = self.window.controller.media.get_mode()
            if media_mode == "video":
                if context.model and context.model.is_video_output():
                    return self.video.generate(
                        context=context,
                        extra=extra,
                    )  # async handled if allowed
            elif media_mode == "image":
                return self.image.generate(
                    context=context,
                    extra=extra,
                )

        # vision
        elif mode == MODE_VISION:
            response = self.vision.send(
                context=context,
                extra=extra,
            )
            used_tokens = self.vision.get_used_tokens()
            self.vision.append_images(ctx)  # append images to ctx if provided

        # assistants
        elif mode == MODE_ASSISTANT:
            # check if assistant is already running and has tools outputs, then submit them, async handled
            if ctx.run_id is not None and len(tools_outputs) > 0:
                self.assistants.worker.tools_submit(
                    ctx,
                    model_id,
                    tools_outputs,  # list of tools outputs
                )
            else:
                # if not running, then send msg and create new assistant run, async handled
                self.assistants.worker.msg_send(
                    ctx,
                    thread_id,
                    assistant_id,
                    model_id,
                    file_ids,
                    prompt,
                    system_prompt,
                )
            return True  # if assistant mode then return here, will be handled async

        # ------- streaming response -------

        if stream:
            ctx.stream = response  # generator
            ctx.set_output("", ai_name)  # set empty output
            ctx.input_tokens = used_tokens  # get from input tokens calculation
            return True  # <-- end here, will be handled in chat stream controller

        # ------- non-streaming response -------

        if response is None:
            return False

        # check for errors
        if "error" in response:
            print("Error in GPT response: " + str(response["error"]))
            return False

        ctx.ai_name = ai_name

        # post-unpack response data, like: output text, tool calls, images, file citations, etc.
        if not use_responses_api:
            self.chat.unpack_response(mode, response, ctx)
        else:
            self.responses.unpack_response(mode, response, ctx)

        return True

    def quick_call(self, context: BridgeContext, extra: dict = None) -> str:
        """
        Quick call OpenAI API with custom prompt

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response content
        """
        # if normal request call then redirect
        if context.request:
            context.stream = False
            context.mode = "chat"  # fake mode for redirect
            self.locked = True
            self.call(context, extra)
            self.locked = False
            return context.ctx.output

        self.locked = True
        ctx = context.ctx
        mode = context.mode
        prompt = context.prompt
        system_prompt = context.system_prompt
        max_tokens = context.max_tokens
        temperature = context.temperature
        functions = context.external_functions
        history = context.history
        model = context.model
        if model is None:
            model = self.window.core.models.from_defaults()

        client = self.get_client(mode, model)
        messages = []
        messages.append({"role": "system", "content": system_prompt})

        if history:
            for item in history:
                messages.append({
                    "role": "user",
                    "content": str(item.final_input)
                })
                messages.append({
                    "role": "assistant",
                    "content": str(item.final_output)
                })
        messages.append({"role": "user", "content": prompt})
        additional_kwargs = {}
        # if max_tokens > 0:
        # additional_kwargs["max_tokens"] = max_tokens

        # tools / functions
        tools = self.window.core.api.openai.tools.prepare(model, functions)
        if len(tools) > 0 and "disable_tools" not in extra:
            additional_kwargs["tools"] = tools

        try:
            response = client.chat.completions.create(
                messages=messages,
                model=model.id,
                temperature=temperature,
                top_p=1.0,
                frequency_penalty=0.0,
                presence_penalty=0.0,
                **additional_kwargs,
            )
            # extract tool calls
            if ctx and response.choices[0].message.tool_calls:
                ctx.tool_calls = self.window.core.command.unpack_tool_calls(
                    response.choices[0].message.tool_calls,
                )
            return response.choices[0].message.content
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error in GPT quick call: " + str(e))
        finally:
            self.locked = False

    def stop(self):
        """On global event stop"""
        pass

    def close(self):
        """Close OpenAI client"""
        if self.locked:
            return
        if self.client is not None:
            try:
                pass
                # self.client.close()
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error closing GPT client:", e)

    def safe_close(self):
        """Close client"""
        if self.locked:
            return
        if self.client is not None:
            try:
                self.client.close()
                self.client = None
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error closing client:", e)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.07 05:00:00                  #
# ================================================== #

def output_html2text(html: str) -> str:
    """
    Convert output HTML to plain text

    :param html: HTML content
    :return: Plain text
    """
    if not html:
        return ""
    try:
        from bs4 import BeautifulSoup, NavigableString
        try:
            soup = BeautifulSoup(html, 'lxml')
        except Exception:
            soup = BeautifulSoup(html, 'html.parser')
        # remove headers from code blocks
        for tag in soup.select('p.code-header-wrapper'):
            tag.replace_with(NavigableString('\n'))
        for tag in soup.select('span.ts, span.toggle-cmd-output, div.name-header'):
            tag.decompose()
        # add separators
        for tag in soup.select('div.msg-bot, div.msg-user'):
            tag.insert_before(NavigableString('\n\n'))
        for br in soup.select('.msg-user .msg p br'):
            br.replace_with('\n')
        text = soup.get_text(separator="", strip=False)
        return text.replace('\t', '    ')
    except Exception:
        pass
    return ""


def output_clean_html(html: str) -> str:
    """
    Clean output HTML content

    :param html: HTML content
    :return: HTML content
    """
    try:
        from bs4 import BeautifulSoup
        try:
            soup = BeautifulSoup(html, 'lxml')
        except Exception:
            soup = BeautifulSoup(html, 'html.parser')
        # remove copy to clipboard from code blocks
        for tag in soup.select('a.code-header-copy'):
            tag.decompose()
        # remove action icons
        for tag in soup.select('div.action-icons'):
            tag.decompose()
        # remove scripts
        for tag in soup.select('script'):
            tag.decompose()
        return str(soup)
    except Exception:
        pass
    return html


def has_unclosed_code_tag(text: str) -> bool:
    """
    Check if HTML content has unclosed code block

    :param text: HTML content
    :return: True if unclosed code block found
    """
    return False
    if not text:
        return False
    return (text.count('```') % 2) != 0

def elide_filename(name_or_path: str, max_len: int = 45, ellipsis: str = "...", keep_dir: bool = False) -> str:
    """
    Elide a long filename by replacing the middle with an ellipsis, preserving the extension.

    Args:
        name_or_path: Filename or full path.
        max_len: Maximum length of the resulting string (including extension and ellipsis).
        ellipsis: Ellipsis text to insert (e.g., "...").
        keep_dir: If True and a path is provided, keep the directory prefix and elide only the basename.
                  If False, operate on the basename only.

    Returns:
        Elided filename (or path if keep_dir=True).
    """
    import os

    if max_len <= 0:
        return name_or_path

    dirpart, base = os.path.split(name_or_path) if keep_dir else ("", os.path.basename(name_or_path))
    stem, ext = os.path.splitext(base)

    # if already short enough
    if len(base) <= max_len:
        return os.path.join(dirpart, base) if keep_dir else base

    # minimal sanity for very small max_len
    min_needed = len(ext) + len(ellipsis) + 2  # at least 1 char head + 1 char tail
    if max_len < min_needed:
        # degrade gracefully: keep first char, ellipsis, last char, and as much ext as fits
        head = stem[:1] if stem else ""
        tail = stem[-1:] if len(stem) > 1 else ""
        # if ext is too long, trim it (rare edge case)
        ext_trim = ext[: max(0, max_len - len(head) - len(ellipsis) - len(tail))]
        out = f"{head}{ellipsis}{tail}{ext_trim}"
        return os.path.join(dirpart, out) if keep_dir else out

    # compute available budget for visible stem parts
    avail = max_len - len(ext) - len(ellipsis)
    # split budget between head and tail (favor head slightly)
    head_len = (avail + 1) // 2
    tail_len = avail - head_len

    # guardrails
    head_len = max(1, head_len)
    tail_len = max(1, tail_len)

    # build elided name
    head = stem[:head_len]
    tail = stem[-tail_len:] if tail_len <= len(stem) else stem
    out = f"{head}{ellipsis}{tail}{ext}"
    return os.path.join(dirpart, out) if keep_dir else out


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.16 22:00:00                  #
# ================================================== #

import asyncio
import hashlib
import re
import shlex
import time
from typing import Dict, List, Tuple, Any, Optional
from urllib.parse import urlparse

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "mcp"
        self.name = "MCP"
        self.description = "Use remote tools via MCP"
        self.prefix = "RemoteTool"
        self.order = 100
        self.use_locale = True
        self.worker = None
        self.config = Config(self)
        self.init_options()

        # Runtime index for quick execution lookup
        self.tools_index: Dict[str, Dict[str, Any]] = {}

        # In-memory discovery cache (per server)
        self._tools_cache: Dict[str, Dict[str, Any]] = {}
        self._last_config_signature: Optional[str] = None

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX
        Build "cmd" entries based on tools discovered from active MCP servers.
        Applies allow/deny per server. Uses cache with TTL.

        :param data: event data dict
        """
        servers: List[dict] = self.get_option_value("servers") or []
        active_servers = [(i, s) for i, s in enumerate(servers) if s.get("active", False)]
        self.tools_index.clear()

        if len(active_servers) == 0:
            return

        # Invalidate cache if config changed
        current_sig = self._config_signature(active_servers)
        if current_sig != self._last_config_signature:
            self._tools_cache.clear()
            self._last_config_signature = current_sig

        try:
            discovered = self._discover_tools_sync(active_servers)
        except Exception as e:
            self.error(e)
            self.error(f"MCP: discovery failed: {e}")
            return

        used_names = set()  # to ensure unique tool names in this batch

        for (server_idx, server_tag, transport, tool, server_cfg) in discovered:
            tool_name = getattr(tool, "name", None) or tool.get("name")
            description = getattr(tool, "description", None) or tool.get("description")
            input_schema = getattr(tool, "inputSchema", None) or tool.get("inputSchema")

            # Human-friendly display name
            display_name = tool_name
            try:
                from mcp.shared.metadata_utils import get_display_name  # type: ignore
                display_name = get_display_name(tool) or tool_name
            except Exception:
                pass

            # Server label -> used in tool name (avoid dots and other invalid chars)
            server_label = (server_cfg.get("label") or server_tag or f"srv{server_idx}").strip()
            server_slug = self._slugify(server_label)

            # Compose final tool "cmd" name acceptable by OpenAI: ^[a-zA-Z0-9_-]+$, <=64 chars
            cmd_name = self._compose_cmd_name(server_slug, tool_name, used_names)
            used_names.add(cmd_name)

            params = self.extract_params_from_schema(input_schema)

            # Instruction for the model
            if description and display_name and display_name != tool_name:
                instruction = f"{display_name}: {description} (server: {server_label})"
            elif description:
                instruction = f"{description} (server: {server_label})"
            else:
                instruction = f"Call remote MCP tool '{display_name}' on server '{server_label}'."

            cmd_syntax = {
                "cmd": cmd_name,
                "instruction": instruction,
                "params": params,
                "enabled": True,
            }
            data['cmd'].append(cmd_syntax)

            # Index for execution
            self.tools_index[cmd_name] = {
                "server_idx": server_idx,
                "server": server_cfg,
                "server_tag": server_tag,
                "transport": transport,
                "tool_name": tool_name,
                "schema": input_schema,
                "description": description,
                "display_name": display_name,
            }

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        from .worker import Worker

        my_commands = [item for item in cmds if item.get("cmd") in self.tools_index]
        if len(my_commands) == 0:
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.plugin = self
            worker.cmds = my_commands
            worker.ctx = ctx
            worker.tools_index = self.tools_index

            if not self.is_async(ctx):
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    # ---------------------------
    # Discovery + caching
    # ---------------------------

    def _discover_tools_sync(self, active_servers: List[Tuple[int, dict]]) -> List[Tuple[int, str, str, Any, dict]]:
        """Run async discovery in a dedicated loop and return collected tools."""
        return asyncio.run(self._discover_tools_async(active_servers))

    async def _discover_tools_async(
        self,
        active_servers: List[Tuple[int, dict]],
        per_server_timeout: float = 8.0
    ) -> List[Tuple[int, str, str, Any, dict]]:
        """
        Discover tools for each active server (with cache).
        Returns tuples: (server_idx, server_tag, transport, tool, server_cfg)
        """
        results: List[Tuple[int, str, str, Any, dict]] = []

        # Lazy import
        try:
            from mcp import ClientSession  # type: ignore
            from mcp.client.stdio import stdio_client  # type: ignore
            from mcp.client.streamable_http import streamablehttp_client  # type: ignore
            from mcp.client.sse import sse_client  # type: ignore
            from mcp import StdioServerParameters  # type: ignore
        except Exception as e:
            self.error('MCP SDK not installed. Install with: pip install "mcp[cli]"')
            self.log(f"MCP import error: {e}")
            return results

        cache_enabled = bool(self.get_option_value("tools_cache_enabled"))
        try:
            ttl = int(self.get_option_value("tools_cache_ttl") or 300)
        except Exception:
            ttl = 300

        for server_idx, server in active_servers:
            address = (server.get("server_address") or "").strip()
            if not address:
                continue

            transport = self._detect_transport(address)
            server_tag = self._make_server_tag(server, server_idx)
            server_key = self._server_key(server)
            headers = self._build_headers(server)

            allowed = self._parse_csv(server.get("allowed_commands"))
            disabled = self._parse_csv(server.get("disabled_commands"))

            # Cache
            cached_tools = None
            if cache_enabled:
                cached = self._tools_cache.get(server_key)
                if cached and cached.get("transport") == transport:
                    if (time.time() - float(cached.get("ts", 0))) <= ttl:
                        cached_tools = cached.get("tools", None)

            async def list_tools_for_session(session: ClientSession) -> List[Any]:
                tools_resp = await session.list_tools()
                return list(tools_resp.tools)

            try:
                if cached_tools is None:
                    async def _run_discovery():
                        if transport == "stdio":
                            cmd, args = self._parse_stdio_command(address)
                            params = StdioServerParameters(command=cmd, args=args)
                            async with stdio_client(params) as (read, write):
                                async with ClientSession(read, write) as session:
                                    await session.initialize()
                                    return await list_tools_for_session(session)

                        elif transport == "http":
                            # Streamable HTTP – pass Authorization if set
                            async with streamablehttp_client(address, headers=headers or None) as (read, write, _):
                                async with ClientSession(read, write) as session:
                                    await session.initialize()
                                    return await list_tools_for_session(session)

                        elif transport == "sse":
                            # SSE – pass Authorization if set
                            async with sse_client(address, headers=headers or None) as (read, write):
                                async with ClientSession(read, write) as session:
                                    await session.initialize()
                                    return await list_tools_for_session(session)

                        else:
                            raise RuntimeError(f"Unsupported MCP transport for server '{server_tag}': {transport}")

                    tools = await asyncio.wait_for(_run_discovery(), timeout=per_server_timeout)

                    if cache_enabled:
                        self._tools_cache[server_key] = {
                            "ts": time.time(),
                            "transport": transport,
                            "tools": tools,
                        }
                else:
                    tools = cached_tools

                for tool in tools:
                    tname = getattr(tool, "name", None) or tool.get("name")
                    if disabled and tname in disabled:
                        continue
                    if allowed and tname not in allowed:
                        continue
                    results.append((server_idx, server_tag, transport, tool, server))

            except asyncio.TimeoutError:
                self.error(f"MCP: timeout during discovery on server '{server_tag}'")
            except Exception as e:
                self.log(f"MCP discovery error on '{server_tag}': {e}")
                self.error(f"MCP: discovery error on '{server_tag}': {e}")

        return results

    # --------------
    # Schema helpers
    # --------------

    def extract_params(self, text: str) -> list:
        """Extract params to list."""
        params = []
        if text is None or text == "":
            return params
        params_list = text.split(",")
        for param in params_list:
            param = param.strip()
            if param == "":
                continue
            params.append({
                "name": param,
                "type": "str",
                "description": param,
            })
        return params

    def extract_params_from_schema(self, schema: Optional[dict]) -> List[dict]:
        """Convert MCP tool inputSchema (JSON Schema) to {name, type, description} list."""
        params: List[dict] = []
        if not schema or not isinstance(schema, dict):
            return params

        properties = schema.get("properties", {})
        required = set(schema.get("required", []) or [])

        for name, prop in properties.items():
            jtype = prop.get("type", "string")
            desc = prop.get("description", "")
            ptype = self._map_json_type_to_param_type(jtype)
            if name in required and desc:
                desc = f"{desc} [required]"
            elif name in required:
                desc = "[required]"
            params.append({
                "name": name,
                "type": ptype,
                "description": desc or name,
            })

        return params

    # ---------------------------
    # Low-level utilities
    # ---------------------------

    def _map_json_type_to_param_type(self, jtype: str) -> str:
        """Map JSON Schema types to simple plugin param types."""
        t = (jtype or "string").lower()
        if t in ("string",):
            return "str"
        if t in ("integer", "number"):
            return "float" if t == "number" else "int"
        if t in ("boolean",):
            return "bool"
        if t in ("array", "object"):
            return "str"
        return "str"

    def _parse_csv(self, text: Optional[str]) -> Optional[set]:
        """Parse comma-separated string into a set of stripped items or None if empty."""
        if not text:
            return None
        items = [x.strip() for x in text.split(",")]
        items = [x for x in items if x]
        return set(items) if items else None

    def _detect_transport(self, address: str) -> str:
        """
        Detect transport from address:
        - 'stdio: ...' -> stdio
        - 'http(s)://.../mcp' or general http(s) -> http (Streamable HTTP)
        - 'sse://' or 'sse+http(s)://' or path containing '/sse' -> sse
        """
        if address.lower().startswith("stdio:"):
            return "stdio"
        lower = address.lower()
        if lower.startswith(("sse://", "sse+http://", "sse+https://")):
            return "sse"
        if lower.startswith(("http://", "https://")):
            try:
                parsed = urlparse(address)
                path = (parsed.path or "").lower()
            except Exception:
                path = ""
            if "/sse" in path or path.endswith("/sse"):
                return "sse"
            return "http"
        return "stdio"

    def _parse_stdio_command(self, address: str) -> Tuple[str, List[str]]:
        """Parse 'stdio: <command line>' into (command, args)."""
        cmdline = address[len("stdio:"):].strip()
        tokens = shlex.split(cmdline)
        if not tokens:
            raise ValueError("Invalid stdio address: empty command")
        return tokens[0], tokens[1:]

    def _make_server_tag(self, server: dict, idx: int) -> str:
        """
        Create a short tag for the server (for display only).
        Prefer explicit 'label'; fallback to host/path-derived value.
        """
        label = (server.get("label") or "").strip()
        if label:
            return label
        address = (server.get("server_address") or "").strip()
        if address.startswith("stdio:"):
            cmdline = address[len("stdio:"):].strip()
            exe = shlex.split(cmdline)[0] if cmdline else f"stdio_{idx}"
            return exe
        try:
            parsed = urlparse(address)
            host = (parsed.netloc or f"server_{idx}")
            tail = (parsed.path.rstrip("/").split("/")[-1] or "mcp")
            return f"{host}_{tail}"
        except Exception:
            return f"server_{idx}"

    def _build_headers(self, server: dict) -> Optional[dict]:
        """
        Build optional headers for HTTP/SSE transports.
        Currently supports Authorization only.
        """
        auth = (server.get("authorization") or "").strip()
        headers = {}
        if auth:
            # If user passed only token, you may expect 'Bearer <token>'
            headers["Authorization"] = auth
        return headers or None

    def _slugify(self, text: str) -> str:
        """
        Sanitize text to allowed chars for tool names: [a-zA-Z0-9_-]
        Collapse multiple underscores and strip from ends.
        """
        if not text:
            return "srv"
        s = re.sub(r"[^a-zA-Z0-9_-]+", "_", text)
        s = re.sub(r"_+", "_", s).strip("_")
        return s or "srv"

    def _truncate_with_hash(self, base: str, max_len: int) -> str:
        """
        Truncate a string to max_len with a short hash suffix to preserve uniqueness.
        """
        if len(base) <= max_len:
            return base
        h = hashlib.sha1(base.encode("utf-8")).hexdigest()[:6]
        keep = max_len - 7  # 1 for '-' + 6 for hash
        keep = max(1, keep)
        return f"{base[:keep]}-{h}"

    def _compose_cmd_name(self, server_slug: str, tool_name: str, used: set) -> str:
        """
        Compose final command name:
        - No global prefixes
        - Format: <server_slug>__<tool_slug>
        - Allowed charset: [a-zA-Z0-9_-]
        - Max length: 64 (OpenAI requirement)
        - Ensure uniqueness within one CMD_SYNTAX build
        """
        tool_slug = self._slugify(tool_name)

        # Initial compose and length guard
        base = f"{server_slug}__{tool_slug}"
        name = self._truncate_with_hash(base, 64)

        # Ensure uniqueness; add numeric suffix if needed (within 64 limit)
        if name not in used:
            return name

        i = 2
        while True:
            suffix = f"-{i}"
            max_len = 64 - len(suffix)
            candidate = self._truncate_with_hash(base, max_len) + suffix
            if candidate not in used:
                return candidate
            i += 1

    def _server_key(self, server: dict) -> str:
        """Deterministic key for a server config entry."""
        addr = (server.get("server_address") or "").strip()
        if addr.lower().startswith("http"):
            try:
                parsed = urlparse(addr)
                return f"http::{parsed.netloc}{parsed.path}"
            except Exception:
                return f"http::{addr}"
        if addr.lower().startswith(("sse://", "sse+http://", "sse+https://")):
            return f"sse::{addr}"
        if addr.startswith("stdio:"):
            return f"stdio::{addr[len('stdio:'):].strip()}"
        return addr

    def _config_signature(self, active_servers: List[Tuple[int, dict]]) -> str:
        """Signature of current config to invalidate cache when config changes."""
        norm: List[str] = []
        for idx, srv in active_servers:
            addr = (srv.get("server_address") or "").strip()
            label = (srv.get("label") or "").strip()
            auth = (srv.get("authorization") or "").strip()
            a = ",".join(sorted(list(self._parse_csv(srv.get("allowed_commands")) or [])))
            d = ",".join(sorted(list(self._parse_csv(srv.get("disabled_commands")) or [])))
            norm.append(f"{idx}|{label}|{addr}|AUTH:{bool(auth)}|A:{a}|D:{d}")
        blob = "|#|".join(norm)
        return hashlib.sha256(blob.encode("utf-8")).hexdigest()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.26 01:00:00                  #
# ================================================== #

from typing import Dict, Any, Tuple, Union, Optional

from agents import (
    Agent as OpenAIAgent,
)

from pygpt_net.core.agents.bridge import ConnectionContext
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.types import (
    AGENT_MODE_OPENAI,
    AGENT_TYPE_OPENAI,
)

from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

from pygpt_net.provider.api.openai.agents.remote_tools import append_tools
from pygpt_net.provider.api.openai.agents.experts import get_experts
from pygpt_net.utils import trans

from ..base import BaseAgent
from .bots.research_bot.manager import ResearchManager


class Agent(BaseAgent):

    PROMPT_PLANNER = (
        "You are a helpful research assistant. Given a query, come up with a set of web searches "
        "to perform to best answer the query. Output between 5 and 20 terms to query for."
    )

    PROMPT_SEARCH = (
        "You are a research assistant. Given a search term, you search the web for that term and "
        "produce a concise summary of the results. The summary must be 2-3 paragraphs and less than 300 "
        "words. Capture the main points. Write succinctly, no need to have complete sentences or good "
        "grammar. This will be consumed by someone synthesizing a report, so its vital you capture the "
        "essence and ignore any fluff. Do not include any additional commentary other than the summary "
        "itself."
    )

    PROMPT_WRITER = (
        "You are a senior researcher tasked with writing a cohesive report for a research query. "
        "You will be provided with the original query, and some initial research done by a research "
        "assistant.\n"
        "You should first come up with an outline for the report that describes the structure and "
        "flow of the report. Then, generate the report and return that as your final output.\n"
        "The final output should be in markdown format, and it should be lengthy and detailed. Aim "
        "for 5-10 pages of content, at least 1000 words."
    )

    def __init__(self, *args, **kwargs):
        super(Agent, self).__init__(*args, **kwargs)
        self.id = "openai_agent_bot_researcher"
        self.type = AGENT_TYPE_OPENAI
        self.mode = AGENT_MODE_OPENAI
        self.name = "Research bot"

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        context = kwargs.get("context", BridgeContext())
        preset = context.preset
        system_prompt = kwargs.get("system_prompt", "")
        agent_name = preset.name if preset else "Agent"
        model = kwargs.get("model", ModelItem())
        tools = kwargs.get("function_tools", [])
        handoffs = kwargs.get("handoffs", [])
        kwargs = {
            "name": agent_name,
            "instructions": system_prompt,
            "model": window.core.agents.provider.get_openai_model(model),
            "handoffs": handoffs,
        }
        tool_kwargs = append_tools(
            tools=tools,
            window=window,
            model=model,
            preset=preset,
            allow_local_tools=True,
            allow_remote_tools=True,
        )
        kwargs.update(tool_kwargs)  # update kwargs with tools
        return OpenAIAgent(**kwargs)

    async def run(
            self,
            window: Any = None,
            agent_kwargs: Dict[str, Any] = None,
            previous_response_id: str = None,
            messages: list = None,
            ctx: CtxItem = None,
            stream: bool = False,
            bridge: ConnectionContext = None,
            use_partial_ctx: Optional[bool] = False,
    ) -> Tuple[CtxItem, str, Union[str, None]]:
        """
        Run agent (async)

        :param window: Window instance
        :param agent_kwargs: Additional agent parameters
        :param previous_response_id: ID of the previous response (if any)
        :param messages: Conversation messages
        :param ctx: Context item
        :param stream: Whether to stream output
        :param bridge: Connection context for agent operations
        :param use_partial_ctx: Use partial ctx per cycle
        :return: Current ctx, final output, last response ID
        """
        response_id = None
        model = agent_kwargs.get("model", ModelItem())
        verbose = agent_kwargs.get("verbose", False)
        context = agent_kwargs.get("context", BridgeContext())
        preset = context.preset if context else None
        query = context.prompt if context else ""
        max_steps = agent_kwargs.get("max_iterations", 10)
        tools = agent_kwargs.get("function_tools", [])

        # allow usage of custom prompt for the writer (final report)
        prompt = self.PROMPT_WRITER
        custom_prompt = self.get_option(preset, "writer", "prompt")
        if custom_prompt and custom_prompt.strip() != "":
            prompt = custom_prompt

        model_planner = window.core.models.get(
            self.get_option(preset, "planner", "model")
        )
        model_search = window.core.models.get(
            self.get_option(preset, "search", "model")
        )

        # prepare provider config
        model_kwargs = {}
        model_search_kwargs = {}
        model_planner_kwargs = {}

        # get experts
        experts = get_experts(
            window=window,
            preset=preset,
            verbose=verbose,
            tools=tools,
        )

        bot = ResearchManager(
            window=window,
            ctx=ctx,
            preset=preset,
            tools=tools,
            bridge=bridge,
            stream=stream,
            planner_config={
                "prompt": self.get_option(preset, "planner", "prompt"),
                "model": model_planner,
                "allow_local_tools": self.get_option(preset, "planner", "allow_local_tools"),
                "allow_remote_tools": self.get_option(preset, "planner", "allow_remote_tools"),
                "run_kwargs": model_planner_kwargs,
                "experts": experts,
            },
            search_config={
                "prompt": self.get_option(preset, "search", "prompt"),
                "model": model_search,
                "allow_local_tools": self.get_option(preset, "search", "allow_local_tools"),
                "allow_remote_tools": self.get_option(preset, "search", "allow_remote_tools"),
                "run_kwargs": model_search_kwargs,
                "experts": experts,
            },
            writer_config={
                "prompt": prompt,
                "model": model,
                "allow_local_tools": self.get_option(preset, "writer", "allow_local_tools"),
                "allow_remote_tools": self.get_option(preset, "writer", "allow_remote_tools"),
                "run_kwargs": model_kwargs,
                "experts": experts,
            },
            history=messages if messages else [],
        )
        final_output = await bot.run(query)
        return ctx, final_output, response_id


    def get_options(self) -> Dict[str, Any]:
        """
        Return Agent options

        :return: dict of options
        """
        return {
            "writer": {
                "label": trans("agent.option.section.base"),
                "options": {
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.base.desc"),
                        "default": self.PROMPT_WRITER,
                    },
                    "allow_local_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.local"),
                        "description": trans("agent.option.tools.local.desc"),
                        "default": False,
                    },
                    "allow_remote_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.remote"),
                        "description": trans("agent.option.tools.remote.desc"),
                        "default": False,
                    },
                }
            },
            "planner": {
                "label": trans("agent.option.section.planner"),
                "options": {
                    "model": {
                        "label": trans("agent.option.model"),
                        "type": "combo",
                        "use": "models",
                        "default": "gpt-4o",
                    },
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.planner.desc"),
                        "default": self.PROMPT_PLANNER,
                    },
                    "allow_local_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.local"),
                        "description": trans("agent.option.tools.local.desc"),
                        "default": False,
                    },
                    "allow_remote_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.remote"),
                        "description": trans("agent.option.tools.remote.desc"),
                        "default": False,
                    },
                }
            },
            "search": {
                "label": trans("agent.option.section.search"),
                "options": {
                    "model": {
                        "label": trans("agent.option.model"),
                        "type": "combo",
                        "use": "models",
                        "default": "gpt-4o",
                    },
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.search.desc"),
                        "default": self.PROMPT_SEARCH,
                    },
                    "allow_local_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.local"),
                        "description": trans("agent.option.tools.local.desc"),
                        "default": False,
                    },
                    "allow_remote_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.remote"),
                        "description": trans("agent.option.tools.remote.desc"),
                        "default": True,
                    },
                }
            },
        }





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.27 06:00:00                  #
# ================================================== #

from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from time import perf_counter

from agents import Runner, TResponseInputItem
from pygpt_net.core.agents.bridge import ConnectionContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem
from pygpt_net.item.preset import PresetItem
from pygpt_net.provider.api.openai.agents.response import StreamHandler

from .logging import Logger, NullLogger
from .schema import FlowSchema, AgentNode, parse_schema
from .graph import FlowGraph, build_graph
from .memory import MemoryManager
from .factory import AgentFactory
from .router import parse_route_output
from .utils import (
    sanitize_input_items,
    extract_text_output,
    patch_last_assistant_output,
    OptionGetter,
    resolve_node_runtime,
)
from .router_streamer import DelayedRouterStreamer, RealtimeRouterStreamer
from .debug import items_preview, ellipsize


@dataclass
class FlowResult:
    ctx: CtxItem
    final_output: str
    last_response_id: Optional[str]


@dataclass
class DebugConfig:
    log_runtime: bool = True
    log_routes: bool = True
    log_inputs: bool = False
    log_outputs: bool = False
    preview_chars: int = 280


class FlowOrchestrator:
    """
    Orchestrates dynamic multi-agent flow based on NodeEditor schema.
    UI semantics follow "bot-to-bot" and supports router stream modes: off/delayed/realtime.
    Memory/no-memory input policy:
      - First agent (in the whole flow) gets full initial messages from the app.
      - Next agent WITHOUT memory gets only last step's displayed content as a single 'user' message.
      - Agent WITH memory:
          * if memory has items -> use base history (items[:-1]) and pass last displayed content as a single 'user' baton;
          * if memory empty -> seed baton from last displayed content (or initial messages as fallback).
    """
    def __init__(self, window, logger: Optional[Logger] = None) -> None:
        self.window = window
        self.logger = logger or NullLogger()

    # ---------- Helpers (production-ready) ----------

    def _extract_text_from_item(self, item: TResponseInputItem) -> str:
        """Best-effort extract plain text from TResponseInputItem."""
        if isinstance(item, dict):
            content = item.get("content", "")
            if isinstance(content, str):
                return content
            if isinstance(content, list):
                parts = []
                for p in content:
                    if isinstance(p, dict):
                        t = p.get("text")
                        if isinstance(t, str):
                            parts.append(t)
                return "\n".join(parts)
            return ""
        if isinstance(item, str):
            return item
        return ""

    def _build_baton_input(
        self,
        *,
        node_id: str,
        g: FlowGraph,
        mem: MemoryManager,
        initial_messages: List[TResponseInputItem],
        first_dispatch_done: bool,
        last_plain_output: str,
        dbg: DebugConfig,
    ) -> tuple[List[TResponseInputItem], str, Optional[str], Any, str]:
        """
        Returns: (prepared_items, baton_user_text, mem_id, mem_state, source_tag)
        Mirrors LI baton/memory policy.
        """
        mem_id = g.agent_to_memory.get(node_id)
        mem_state = mem.get(mem_id) if mem_id else None

        baton_user_text = ""
        source = ""

        if mem_state and mem_state.items:
            # memory with history -> base history + baton from last output (preferred)
            base_items = list(mem_state.items[:-1]) if len(mem_state.items) >= 1 else []
            if last_plain_output and last_plain_output.strip():
                baton_user_text = last_plain_output
                prepared = base_items + [{"role": "user", "content": baton_user_text}]
                source = "memory:existing_to_user_baton"
            else:
                # fallback: use last assistant content as baton
                last_ass = mem_state.items[-1] if isinstance(mem_state.items[-1], dict) else {}
                if isinstance(last_ass.get("content"), str):
                    baton_user_text = last_ass.get("content", "")
                elif isinstance(last_ass.get("content"), list) and last_ass["content"]:
                    baton_user_text = last_ass["content"][0].get("text", "") or ""
                else:
                    baton_user_text = ""
                prepared = base_items + [{"role": "user", "content": baton_user_text}]
                source = "memory:existing_to_last_assistant"
            return sanitize_input_items(prepared), baton_user_text, mem_id, mem_state, source

        if mem_state:
            # memory attached but empty -> seed from last output else from initial (use last user msg as baton)
            if last_plain_output and last_plain_output.strip():
                baton_user_text = last_plain_output
                prepared = [{"role": "user", "content": baton_user_text}]
                source = "memory:seed_from_last_output"
            else:
                base_items = list(initial_messages[:-1]) if initial_messages else []
                last_item = initial_messages[-1] if initial_messages else {"role": "user", "content": ""}
                baton_user_text = self._extract_text_from_item(last_item)
                prepared = base_items + [{"role": "user", "content": baton_user_text}]
                source = "memory:seed_from_initial"
            return sanitize_input_items(prepared), baton_user_text, mem_id, mem_state, source

        # no memory attached
        if not first_dispatch_done:
            # first agent: pass initial messages as-is; baton is last user text (for potential external memory)
            last_item = initial_messages[-1] if initial_messages else {"role": "user", "content": ""}
            baton_user_text = self._extract_text_from_item(last_item)
            return sanitize_input_items(list(initial_messages)), baton_user_text, None, None, "no-mem:first_initial"
        else:
            baton_user_text = last_plain_output if last_plain_output and last_plain_output.strip() else (
                self._extract_text_from_item(initial_messages[-1]) if initial_messages else ""
            )
            prepared = [{"role": "user", "content": baton_user_text}]
            return sanitize_input_items(prepared), baton_user_text, None, None, "no-mem:last_output"

    def _update_memory_after_step(
        self,
        *,
        node_id: str,
        mem_state: Any,
        baton_user_text: str,
        display_text: str,
        last_response_id: Optional[str],
        dbg: DebugConfig,
    ) -> None:
        """Update memory strictly with [user baton, assistant display_text], mirroring LI semantics."""
        if not mem_state:
            return
        base_items = list(mem_state.items[:-1]) if getattr(mem_state, "items", None) else []
        new_mem = (base_items or []) + [
            {"role": "user", "content": baton_user_text or ""},
            {"role": "assistant", "content": [{"type": "output_text", "text": display_text or ""}]},
        ]
        try:
            mem_state.set_from(new_mem, last_response_id)
            if dbg.log_inputs:
                self.logger.debug(
                    f"[memory] {node_id} updated len {len(base_items)} -> {len(new_mem)} "
                    f"user='{ellipsize(baton_user_text or '', dbg.preview_chars)}' "
                    f"assist='{ellipsize(display_text or '', dbg.preview_chars)}'"
                )
        except Exception as e:
            self.logger.error(f"[memory] update failed for {node_id}: {e}")

    # ---------- Main flow ----------

    async def run_flow(
        self,
        schema: List[Dict[str, Any]],
        messages: List[TResponseInputItem],
        ctx: CtxItem,
        bridge: ConnectionContext,
        agent_kwargs: Dict[str, Any],
        preset: Optional[PresetItem],
        model: ModelItem,
        stream: bool,
        use_partial_ctx: bool,
        base_prompt: Optional[str],
        allow_local_tools_default: bool,
        allow_remote_tools_default: bool,
        function_tools: List[dict],
        trace_id: Optional[str],
        max_iterations: int = 20,
        router_stream_mode: str = "off",  # "off" | "delayed" | "realtime"
        option_get: Optional[OptionGetter] = None,
    ) -> FlowResult:
        fs: FlowSchema = parse_schema(schema)
        g: FlowGraph = build_graph(fs)
        mem = MemoryManager()
        factory = AgentFactory(self.window, self.logger)
        option_get = option_get or (lambda s, k, d=None: d)

        # Debug config
        dbg = DebugConfig(
            log_runtime=bool(option_get("debug", "log_runtime", True)),
            log_routes=bool(option_get("debug", "log_routes", True)),
            log_inputs=bool(option_get("debug", "log_inputs", True)),
            log_outputs=bool(option_get("debug", "log_outputs", True)),
            preview_chars=int(option_get("debug", "preview_chars", 280)),
        )

        # Entry
        if g.start_targets:
            current_ids: List[str] = [g.start_targets[0]]
            self.logger.info(f"Using explicit START -> {current_ids[0]}")
        else:
            default_agent = g.pick_default_start_agent()
            if default_agent is None:
                self.logger.error("No START and no agents in schema.")
                return FlowResult(ctx=ctx, final_output="", last_response_id=None)
            current_ids = [default_agent]
            self.logger.info(f"No START found, using lowest-id agent: {default_agent}")

        # State
        final_output = ""
        last_response_id: Optional[str] = None

        # Initial messages for the very first agent in flow
        initial_messages: List[TResponseInputItem] = sanitize_input_items(list(messages or []))
        first_dispatch_done: bool = False
        last_plain_output: str = ""  # what was displayed to UI in the previous step

        steps = 0

        # Shared stream handler (bot-to-bot style)
        handler = StreamHandler(self.window, bridge)
        begin = True

        while current_ids and (steps < max_iterations or max_iterations == 0) and not bridge.stopped():
            step_start = perf_counter()
            current_id = current_ids[0]
            steps += 1

            # END node
            if current_id in fs.ends:
                self.logger.info(f"Reached END node: {current_id}")
                break

            # Validate agent
            if current_id not in fs.agents:
                self.logger.warning(f"Next id {current_id} is not an agent; stopping or jumping to END.")
                if g.end_nodes:
                    current_ids = [g.end_nodes[0]]
                    continue
                break

            node: AgentNode = fs.agents[current_id]
            self.logger.debug(f"[step {steps}] agent_id={node.id} name={node.name} outs={node.outputs}")

            # Resolve per-node runtime
            node_rt = resolve_node_runtime(
                window=self.window,
                node=node,
                option_get=option_get,
                default_model=model,
                base_prompt=base_prompt,
                schema_allow_local=node.allow_local_tools,
                schema_allow_remote=node.allow_remote_tools,
                default_allow_local=allow_local_tools_default,
                default_allow_remote=allow_remote_tools_default,
            )

            if dbg.log_runtime:
                instr_preview = ellipsize(node_rt.instructions, dbg.preview_chars)
                self.logger.debug(
                    f"[runtime] model={getattr(node_rt.model,'name',str(node_rt.model))} "
                    f"allow_local={node_rt.allow_local_tools} allow_remote={node_rt.allow_remote_tools} "
                    f"instructions='{instr_preview}'"
                    f" role='{node_rt.role}'"
                )

            # Input build using baton policy (LI parity)
            prepared_items, baton_user_text, mem_id, mem_state, input_source = self._build_baton_input(
                node_id=current_id,
                g=g,
                mem=mem,
                initial_messages=initial_messages,
                first_dispatch_done=first_dispatch_done,
                last_plain_output=last_plain_output,
                dbg=dbg,
            )

            if dbg.log_inputs:
                self.logger.debug(f"[input] source={input_source} items={len(prepared_items)} "
                                  f"preview={items_preview(prepared_items, dbg.preview_chars)}")
                if mem_id:
                    mem_info = f"{mem_id} (len={len(mem_state.items) if mem_state else 0})"
                    self.logger.debug(f"[memory] attached={bool(mem_id)} mem_id={mem_info}")

            # Build agent with per-node runtime
            # Restrict friendly_map only to allowed outgoing routes of current node
            allowed_map = {rid: fs.agents[rid].name or rid for rid in (node.outputs or []) if rid in fs.agents}
            built = factory.build(
                node=node,
                node_runtime=node_rt,
                preset=preset,
                function_tools=function_tools,
                force_router=False,  # auto on multi-output
                friendly_map=allowed_map,
                handoffs_enabled=True,
                context=agent_kwargs.get("context"),
            )
            agent = built.instance
            multi_output = built.multi_output
            allowed_routes = built.allowed_routes

            if dbg.log_runtime and multi_output:
                self.logger.debug(f"[routing] multi_output=True routes={allowed_routes} mode={router_stream_mode}")

            # Prepare run kwargs
            run_kwargs: Dict[str, Any] = {
                "input": prepared_items,
                "max_turns": int(agent_kwargs.get("max_iterations", max_iterations)),
            }
            # if trace_id:
                # run_kwargs["trace_id"] = trace_id

            # Header for UI
            ctx.set_agent_name(agent.name)
            bridge.on_step(ctx, begin)
            begin = False
            handler.begin = begin

            display_text = ""  # what we show to UI for this step
            next_id: Optional[str] = None

            # --- EXECUTION ---
            if stream and not multi_output:
                # Full token streaming (single-output agent) – collect full buffer for baton
                result = Runner.run_streamed(agent, **run_kwargs)
                handler.reset()
                # Optional local accumulator; prefer handler.buffer after loop
                last_chunk = ""

                async for event in result.stream_events():
                    if bridge.stopped():
                        result.cancel()
                        bridge.on_stop(ctx)
                        break
                    chunk, last_response_id = handler.handle(event, ctx)
                    if chunk:
                        last_chunk = chunk

                # Use full buffer if available (ensures baton sees complete output)
                display_text = getattr(handler, "buffer", "") or last_chunk or ""

                # Update memory strictly with baton + displayed text
                self._update_memory_after_step(
                    node_id=current_id,
                    mem_state=mem_state,
                    baton_user_text=baton_user_text,
                    display_text=display_text,
                    last_response_id=last_response_id,
                    dbg=dbg,
                )

                # Route: first edge or END
                outs = g.get_next(current_id)
                next_id = outs[0] if outs else g.first_connected_end(current_id)

            else:
                # Multi-output or non-stream
                if multi_output:
                    mode = (router_stream_mode or "off").lower()
                    if stream and mode == "realtime":
                        # Realtime router streaming: stream content as tokens arrive
                        result = Runner.run_streamed(agent, **run_kwargs)
                        rts = RealtimeRouterStreamer(
                            window=self.window,
                            bridge=bridge,
                            handler=handler if not use_partial_ctx else None,
                            buffer_to_handler=(not use_partial_ctx),
                            logger=self.logger,
                        )
                        rts.reset()

                        async for event in result.stream_events():
                            if bridge.stopped():
                                result.cancel()
                                bridge.on_stop(ctx)
                                break
                            rts.handle_event(event, ctx)
                            if rts.last_response_id:
                                last_response_id = rts.last_response_id

                        raw_text = rts.buffer or ""
                        decision = parse_route_output(raw_text, allowed_routes)
                        display_text = decision.content or ""

                        # Update memory with baton + displayed content
                        self._update_memory_after_step(
                            node_id=current_id,
                            mem_state=mem_state,
                            baton_user_text=baton_user_text,
                            display_text=display_text,
                            last_response_id=last_response_id,
                            dbg=dbg,
                        )

                        # Route decision
                        if decision.valid:
                            next_id = decision.route
                        else:
                            if dbg.log_routes:
                                self.logger.warning("[router-realtime] Invalid JSON; fallback to first route.")
                            next_id = allowed_routes[0] if allowed_routes else None

                    elif stream and mode == "delayed":
                        # Delayed router streaming: collect tokens silently, reveal once
                        result = Runner.run_streamed(agent, **run_kwargs)
                        delayed = DelayedRouterStreamer(self.window, bridge)
                        delayed.reset()

                        async for event in result.stream_events():
                            if bridge.stopped():
                                result.cancel()
                                bridge.on_stop(ctx)
                                break
                            _, rid = delayed.handle_event(event, ctx)
                            if rid:
                                last_response_id = rid

                        raw_text = delayed.buffer or ""
                        decision = parse_route_output(raw_text, allowed_routes)
                        display_text = decision.content or ""
                        if display_text:
                            ctx.stream = display_text
                            bridge.on_step(ctx, False)
                            if not use_partial_ctx:
                                handler.to_buffer(display_text)

                        # Update memory with baton + displayed content
                        self._update_memory_after_step(
                            node_id=current_id,
                            mem_state=mem_state,
                            baton_user_text=baton_user_text,
                            display_text=display_text,
                            last_response_id=last_response_id,
                            dbg=dbg,
                        )

                        if decision.valid:
                            next_id = decision.route
                        else:
                            if dbg.log_routes:
                                self.logger.warning(f"[router-delayed] Invalid JSON: {decision.error}; fallback first route.")
                            next_id = allowed_routes[0] if allowed_routes else None
                    else:
                        # No streaming for router: run to completion, then emit only content
                        result = await Runner.run(agent, **run_kwargs)
                        last_response_id = getattr(result, "last_response_id", None)
                        raw_text = extract_text_output(result)
                        decision = parse_route_output(raw_text, allowed_routes)
                        display_text = decision.content or ""
                        if display_text:
                            ctx.stream = display_text
                            bridge.on_step(ctx, False)
                            if not use_partial_ctx:
                                handler.to_buffer(display_text)

                        # Update memory with baton + displayed content
                        self._update_memory_after_step(
                            node_id=current_id,
                            mem_state=mem_state,
                            baton_user_text=baton_user_text,
                            display_text=display_text,
                            last_response_id=last_response_id,
                            dbg=dbg,
                        )

                        if decision.valid:
                            next_id = decision.route
                        else:
                            if dbg.log_routes:
                                self.logger.warning(f"[router-off] Invalid JSON: {decision.error}; fallback first route.")
                            next_id = allowed_routes[0] if allowed_routes else None
                else:
                    # Single-output, non-stream path
                    result = await Runner.run(agent, **run_kwargs)
                    last_response_id = getattr(result, "last_response_id", None)
                    raw_text = extract_text_output(result)
                    display_text = raw_text or ""
                    if display_text:
                        ctx.stream = display_text
                        bridge.on_step(ctx, False)
                        if not use_partial_ctx:
                            handler.to_buffer(display_text)

                    # Update memory with baton + displayed text
                    self._update_memory_after_step(
                        node_id=current_id,
                        mem_state=mem_state,
                        baton_user_text=baton_user_text,
                        display_text=display_text,
                        last_response_id=last_response_id,
                        dbg=dbg,
                    )

                    outs = g.get_next(current_id)
                    next_id = outs[0] if outs else g.first_connected_end(current_id)

            # DEBUG: output + route
            if dbg.log_outputs:
                self.logger.debug(f"[output] preview='{ellipsize(display_text, dbg.preview_chars)}' "
                                  f"last_response_id={last_response_id}")

            if dbg.log_routes:
                self.logger.debug(f"[route] current={current_id} -> next={next_id} "
                                  f"(end_connected={g.first_connected_end(current_id)})")

            # Mark dispatch and remember last plain output (for next no-memory agent)
            first_dispatch_done = True
            last_plain_output = display_text
            final_output = display_text

            # Resolve next id / END
            if isinstance(next_id, str) and next_id.lower() == "end":
                end_id = g.first_connected_end(current_id) or (g.end_nodes[0] if g.end_nodes else None)
                current_ids = [end_id] if end_id else []
            elif next_id:
                current_ids = [next_id]
            else:
                end_id = g.first_connected_end(current_id)
                current_ids = [end_id] if end_id else []

            # UI separation after each agent step
            is_end = True if not current_ids else (current_ids[0] in fs.ends)
            if use_partial_ctx:
                ctx = bridge.on_next_ctx(
                    ctx=ctx,
                    input="",
                    output=display_text,
                    response_id=last_response_id or "",
                    finish=is_end,
                    stream=True,
                )
                handler.new()
            else:
                bridge.on_next(ctx)

            # set next agent name if not at the end
            if current_ids and current_ids[0] in fs.agents:
                ctx.set_agent_name(fs.agents[current_ids[0]].name)

            # Step duration
            dur = perf_counter() - step_start
            self.logger.debug(f"[step {steps}] duration={dur:.3f}s")

        if bridge.stopped():
            bridge.on_stop(ctx)

        self.logger.info(f"Flow finished. steps={steps} final_len={len(final_output)}")
        return FlowResult(ctx=ctx, final_output=final_output, last_response_id=last_response_id)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.25 12:00:00                  #
# ================================================== #

from typing import Optional, Any, Dict

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.bridge.context import MultimodalContext
from pygpt_net.core.events import Event, AppEvent, KernelEvent, RenderEvent
from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_ASSISTANT,
    MODE_IMAGE,
)
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Input:
    def __init__(self, window=None):
        """
        Input controller

        :param window: Window instance
        """
        self.window = window
        self.locked = False
        self.stop = False
        self.generating = False
        self.no_ctx_idx_modes = [
            # MODE_IMAGE,
            MODE_ASSISTANT,
            # MODE_LLAMA_INDEX,
            MODE_AGENT,
        ]  # assistant is handled in async, agent is handled in agent flow
        self.stop_commands = [
            "stop",
            "halt",
        ]

    def send_input(self, force: bool = False):
        """
        Send text from user input (called from UI)

        :param force: force send
        """
        dispatch = self.window.dispatch
        mode = self.window.core.config.get('mode')
        event = Event(Event.INPUT_BEGIN, {
            'mode': mode,
            'force': force,
            'stop': False,
        })
        dispatch(event)
        stop = event.data.get('stop', False)

        # get text from input
        text = self.window.ui.nodes['input'].toPlainText().strip()

        if not force:
            dispatch(AppEvent(AppEvent.INPUT_SENT))  # app event
            if stop:
                return

        # listen for stop command
        if self.generating \
                and text is not None \
                and text.lower().strip() in self.stop_commands:
            self.window.controller.kernel.stop()  # TODO: to chat main
            dispatch(RenderEvent(RenderEvent.CLEAR_INPUT))
            return

        # event: user input send (manually)
        event = Event(Event.USER_SEND, {
            'mode': mode,
            'value': text,
        })
        dispatch(event)
        text = event.data['value']

        # if attachments, return here - send will be handled via signal after upload
        if self.handle_attachment(mode, text):
            return

        # kernel event: handle input
        context = BridgeContext()
        context.prompt = text
        dispatch(KernelEvent(KernelEvent.INPUT_USER, {
            'context': context,
            'extra': {},
        }))

    def send(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
    ):
        """
        Send input wrapper

        :param context: bridge context
        :param extra: extra data
        """
        self.execute(
            text=str(context.prompt),
            force=extra.get("force", False),
            reply=extra.get("reply", False),
            internal=extra.get("internal", False),
            prev_ctx=context.ctx,
            multimodal_ctx=context.multimodal_ctx,
        )

    def execute(
            self,
            text: str,
            force: bool = False,
            reply: bool = False,
            internal: bool = False,
            prev_ctx: Optional[CtxItem] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ):
        """
        Execute send input text to API

        :param text: input text
        :param force: force send (ignore input lock)
        :param reply: reply mode (from plugins)
        :param internal: internal call
        :param prev_ctx: previous context (if reply)
        :param multimodal_ctx: multimodal context
        """
        core = self.window.core
        controller = self.window.controller
        dispatch = self.window.dispatch
        log = controller.chat.log

        dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "chat",
        }))

        # check if input is not locked
        if self.locked and not force and not internal:
            return

        log("Begin.")
        self.generating = True  # set generating flag

        # check if assistant is selected
        mode = core.config.get('mode')
        if mode == MODE_ASSISTANT:
            if not controller.assistant.check():
                self.generating = False  # unlock
                return

        # handle camera capture
        controller.camera.handle_auto_capture(mode)

        # unlock if locked
        controller.assistant.resume()
        controller.kernel.resume()

        log(f"Input prompt: {text}")  # log

        # event: before input handle
        event = Event(Event.INPUT_BEFORE, {
            'mode': mode,
            'value': text,
            'multimodal_ctx': multimodal_ctx,
            'stop': False,
            'silent': False,  # silent mode (without error messages)
        })
        dispatch(event)
        text = event.data['value']
        stop = event.data.get('stop', False)
        silent = event.data.get('silent', False)

        if stop:  # abort via event
            self.generating = False
            if not silent:
                dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
                    "id": "chat",
                }))
            return

        # set state to: busy
        dispatch(KernelEvent(KernelEvent.STATE_BUSY, {
            "id": "chat",
            "msg": trans('status.sending'),
        }))

        # clear input field if clear-on-send is enabled
        if core.config.get('send_clear') and not force and not internal:
            dispatch(RenderEvent(RenderEvent.CLEAR_INPUT))

        # create ctx, handle allowed, etc.
        dispatch(Event(Event.INPUT_ACCEPT, {
            'value': text,
            'multimodal_ctx': multimodal_ctx,
            'mode': mode,
        }))

        # start timer
        self.window.controller.chat.common.start_counter()

        # send input to API
        if mode == MODE_IMAGE:
            controller.chat.image.send(
                text=text,
                prev_ctx=prev_ctx,
            )  # image generation
        else:
            controller.chat.text.send(
                text=text,
                reply=reply,
                internal=internal,
                prev_ctx=prev_ctx,
                multimodal_ctx=multimodal_ctx,
            )  # text mode: OpenAI, LlamaIndex, etc.

    def handle_attachment(self, mode: str, text: str) -> bool:
        """
        Handle attachments with additional context (not images here)

        :param mode: Mode (e.g., MODE_ASSISTANT, MODE_CHAT)
        :param text: Input text
        :return: bool: True if attachments exists, False otherwise
        """
        controller = self.window.controller
        dispatch = self.window.dispatch
        exists = False

        # handle attachments with additional context (not images here)
        if mode != MODE_ASSISTANT and controller.chat.attachment.has(mode):
            exists = True
            dispatch(KernelEvent(KernelEvent.STATE_BUSY, {
                "id": "chat",
                "msg": "Reading attachments..."
            }))
            try:
                controller.chat.attachment.handle(mode, text)
            except Exception as e:
                dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
                    "id": "chat",
                    "msg": f"Error reading attachments: {e}"
                }))

        return exists



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.04 19:00:00                  #
# ================================================== #

from typing import Union

from pygpt_net.item.model import ModelItem


class RemoteTools:
    def __init__(self, window=None):
        """
        Remote tools controller

        :param window: Window instance
        """
        self.window = window
        self.enabled_global = {
            "web_search": False,
        }

    def setup(self):
        """
        Setup remote tools

        :return: None
        """
        cfg_get = self.window.core.config.get
        self.enabled_global["web_search"] = cfg_get("remote_tools.global.web_search", False)
        self.update_icons()

    def enabled(self, model: Union[ModelItem, str], tool_name: str) -> bool:
        """
        Check if remote tool is enabled

        :param model: ModelItem or model name
        :param tool_name: Tool name
        :return: True if enabled, False otherwise
        """
        if isinstance(model, str):
            model = self.window.core.models.get(model)
        if not model:
            return False
        if tool_name == "web_search":
            return self.is_web(model)
        return False

    def is_web(self, model: ModelItem) -> bool:
        """
        Check if web search is enabled for the given provider

        :param model: ModelItem
        :return: True if web search is enabled, False otherwise
        """
        # at first, check provider-specific config
        cfg_get = self.window.core.config.get
        state = False
        if model.provider == "openai":  # native SDK, responses API
            state = cfg_get("remote_tools.web_search", False)
        elif model.provider == "google":  # native SDK
            state = cfg_get("remote_tools.google.web_search", False)
        elif model.provider == "anthropic":  # native SDK
            state = cfg_get("remote_tools.anthropic.web_search", False)
        elif model.provider == "x_ai":  # native SDK
            state = cfg_get("remote_tools.xai.web_search", False)

        # if not enabled by default or other provider, then use global config
        if not state:
            state = self.enabled_global["web_search"]

        return state

    def update_icons(self):
        """
        Update remote tools icons in chat tabs
        """
        state = self.enabled_global["web_search"]
        if state:
            self.window.ui.nodes['input'].set_icon_state("web", True)
            # self.window.ui.nodes['icon.remote_tool.web'].set_icon(":/icons/web_on.svg")
        else:
            self.window.ui.nodes['input'].set_icon_state("web", False)
            # self.window.ui.nodes['icon.remote_tool.web'].set_icon(":/icons/web_off.svg")

    def toggle(self, tool_name: str):
        """
        Toggle remote tool (for global toggle button)

        :param tool_name: Tool name
        """
        cfg_set = self.window.core.config.set

        # tool: web search
        if tool_name == "web_search":
            state = not self.enabled_global["web_search"]
            self.enabled_global["web_search"] = state
            cfg_set("remote_tools.global.web_search", state)
            cfg_set("remote_tools.web_search", state)
            cfg_set("remote_tools.google.web_search", state)
            cfg_set("remote_tools.anthropic.web_search", state)
            cfg_set("remote_tools.xai.web_search", state)

        # save config
        self.window.core.config.save()
        self.update_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtWidgets import QApplication

from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.utils import trans, freeze_updates


class Settings:
    def __init__(self, window=None):
        """
        Plugin settings controller

        :param window: Window instance
        """
        self.window = window
        self.config_dialog = False
        self.config_initialized = False
        self.current_plugin = None
        self.width = 800
        self.height = 500

    def setup(self):
        """Set up plugin settings"""
        idx = None
        if 'plugin.settings' in self.window.ui.tabs:
            idx = self.window.ui.tabs['plugin.settings'].currentIndex()
        self.window.plugin_settings.setup(idx)

    def toggle_editor(self):
        """Toggle plugin settings dialog"""
        if self.config_dialog:
            self.close()
        else:
            self.open()

    def open(self):
        """Open plugin settings dialog"""
        if not self.config_initialized:
            self.window.dispatch(KernelEvent(KernelEvent.STATUS, {
                'status': trans("status.loading")
            }))
            QApplication.processEvents()
            with freeze_updates(self.window):
                self.setup()
            self.config_initialized = True
        if self.config_dialog:
            return
        self.init()
        self.window.ui.dialogs.open(
            'plugin_settings',
            width=self.width,
            height=self.height
        )
        self.window.dispatch(KernelEvent(KernelEvent.STATUS, {
            'status': ""
        }))
        self.config_dialog = True

    def open_plugin(self, id: str):
        """
        Open plugin settings dialog

        :param id: plugin id
        """
        self.current_plugin = id
        self.open()

    def init(self):
        """Initialize plugin settings options"""
        plugins = self.window.core.plugins.plugins
        if self.current_plugin is None and plugins:
            self.current_plugin = next(iter(plugins))
        cfg = self.window.controller.config
        for pid, plugin in plugins.items():
            options = plugin.setup()
            cfg.load_options(f'plugin.{pid}', options)
        self.window.controller.layout.restore_plugin_settings()

    def refresh_option(self, id: str, key: str):
        """
        Refresh plugin option

        :param id: plugin id
        :param key: option key
        """
        plugins = self.window.core.plugins.plugins
        plugin = plugins.get(id)
        if not plugin:
            return
        options = plugin.options
        if key not in options:
            return
        option = options[key]
        self.window.controller.config.placeholder.apply(option)
        items = option.get("keys", {})
        self.window.controller.config.update_list(
            option=option,
            parent_id=f'plugin.{id}',
            key=key,
            items=items,
        )

    def save(self):
        """Save plugin settings"""
        window = self.window
        plugins = window.core.plugins.plugins
        controller_cfg = window.controller.config
        config_data = window.core.config.data
        plugins_cfg = config_data.setdefault('plugins', {})

        for pid, plugin in plugins.items():
            options = plugin.setup()
            dst = plugins_cfg.setdefault(pid, {})
            for key, opt in options.items():
                value = controller_cfg.get_value(
                    parent_id=f'plugin.{pid}',
                    key=key,
                    option=opt,
                )
                plugin.options[key]['value'] = value
                dst[key] = value

        stale = set(plugins_cfg.keys()) - set(plugins.keys())
        for pid in stale:
            plugins_cfg.pop(pid, None)

        window.controller.plugins.presets.save_current()
        window.core.config.save()
        self.close()

        window.update_status(trans('info.settings.saved'))
        window.dispatch(Event(Event.PLUGIN_SETTINGS_CHANGED))
        window.controller.ui.update_tokens()

    def close(self):
        """Close plugin settings dialog"""
        if self.config_dialog:
            self.window.ui.dialogs.close('plugin_settings')
            self.config_dialog = False

    def load_defaults_user(self, force: bool = False):
        """
        Load plugin settings user defaults

        :param force: force load defaults
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='plugin.settings.defaults.user',
                id=-1,
                msg=trans('dialog.plugin.settings.defaults.user.confirm'),
            )
            return
        self.init()

    def load_defaults_app(self, force: bool = False):
        """
        Load plugin settings app defaults

        :param force: force load defaults
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='plugin.settings.defaults.app',
                id=-1,
                msg=trans('dialog.plugin.settings.defaults.app.confirm'),
            )
            return
        self.window.core.plugins.restore_options(self.current_plugin)
        self.init()
        self.window.ui.dialogs.alert(trans('dialog.plugin.settings.defaults.app.result'))

    def get_option(self, id: str, key: str) -> Any:
        """
        Get plugin option

        :param id: plugin id
        :param key: option key
        :return: option value
        """
        return self.window.core.plugins.plugins[id].options[key]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.23 15:00:00                  #
# ================================================== #

from typing import List, Optional

from PySide6.QtCore import QRunnable, QObject, Signal, Slot
from llama_index.core.tools import QueryEngineTool

from pygpt_net.core.types import (
    MODE_EXPERT,
    TOOL_EXPERT_CALL_NAME,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent, RenderEvent
from pygpt_net.item.ctx import CtxItem


class WorkerSignals(QObject):
    """Signals for worker to communicate with main thread."""
    finished = Signal()  # when worker is finished
    response = Signal(object, str)  # when worker has response
    error = Signal(str)  # when worker has error
    event = Signal(object)  # when worker has event to dispatch
    output = Signal(object, str)  # when worker has output to handle
    lock_input = Signal()  # when worker locks input for UI
    cmd = Signal(object, object, str, str, str)  # when worker has command to handle


class ExpertWorker(QRunnable):
    """Worker for handling expert calls in a separate thread."""

    def __init__(
            self,
            window,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        super().__init__()
        self.window = window
        self.master_ctx = master_ctx
        self.expert_id = expert_id
        self.query = query
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        master_ctx = self.master_ctx
        expert_id = self.expert_id
        query = self.query

        try:
            # get or create children (slave) meta
            slave = self.window.core.ctx.get_or_create_slave_meta(master_ctx, expert_id)
            expert = self.window.core.experts.get_expert(expert_id)  # preset
            reply = True
            hidden = False
            internal = False

            if self.window.core.experts.agent_enabled():  # hide in agent mode
                internal = False
                hidden = True

            mode = self.window.core.config.get("mode")
            base_mode = mode
            model = expert.model
            expert_name = expert.name
            ai_name = ""
            sys_prompt = expert.prompt
            model_data = self.window.core.models.get(model)

            files = []
            file_ids = []
            functions = []
            tools_outputs = []

            # from current config
            max_tokens = self.window.core.config.get('max_output_tokens')
            stream_mode = self.window.core.config.get('stream')
            verbose = self.window.core.config.get('agent.llama.verbose')
            use_agent = self.window.core.config.get('experts.use_agent', False)
            db_idx = expert.idx  # get idx from expert preset

            mode = MODE_EXPERT  # force expert mode, mode will change in bridge

            # create slave item
            ctx = CtxItem()
            ctx.meta = slave  # use slave-meta
            ctx.internal = internal
            ctx.hidden = hidden
            ctx.current = True  # mark as current context item
            ctx.mode = mode  # store current selected mode (not inline changed)
            ctx.model = model  # store model list key, not real model id
            ctx.set_input(query, str(ai_name))
            ctx.set_output(None, expert_name)
            ctx.sub_call = True  # mark as sub-call
            ctx.pid = master_ctx.pid  # copy PID from parent to allow reply

            # render: begin
            event = RenderEvent(RenderEvent.BEGIN, {
                "meta": ctx.meta,
                "ctx": ctx,
                "stream": stream_mode,
            })
            self.signals.event.emit(event)  # dispatch render event
            self.window.core.ctx.provider.append_item(slave, ctx)  # to slave meta

            # build sys prompt
            sys_prompt_raw = sys_prompt  # store raw prompt
            event = Event(Event.PRE_PROMPT, {
                'mode': mode,
                'value': sys_prompt,
                'is_expert': True,
            })
            self.signals.event.emit(event)  # dispatch pre-prompt event
            sys_prompt = event.data['value']
            sys_prompt = self.window.core.prompt.prepare_sys_prompt(
                mode,
                model_data,
                sys_prompt,
                ctx,
                reply,
                internal,
                is_expert=True,  # mark as expert, blocks expert prompt append in plugin
            )

            # index to use
            use_index = False
            if self.window.core.idx.is_valid(db_idx):
                use_index = True
                self.window.core.experts.last_idx = db_idx  # store last index used in call
            else:
                self.window.core.experts.last_idx = None
            if use_index:
                index, llm = self.window.core.idx.chat.get_index(db_idx, model_data, stream=False)
            else:
                llm = self.window.core.idx.llm.get(model_data, stream=False)

            history = self.window.core.ctx.all(
                meta_id=slave.id
            )  # get history for slave ctx, not master ctx

            if use_agent:
                # call the agent (planner) with tools and index
                ctx.agent_call = True  # directly return tool call response
                ctx.use_agent_final_response = True  # use agent final response as output
                bridge_context = BridgeContext(
                    ctx=ctx,
                    history=history,
                    mode=mode,
                    parent_mode=base_mode,
                    model=model_data,
                    system_prompt=sys_prompt,
                    system_prompt_raw=sys_prompt_raw,
                    prompt=query,
                    stream=False,
                    attachments=files,
                    file_ids=file_ids,
                    assistant_id=self.window.core.config.get('assistant'),
                    idx=db_idx,
                    idx_mode=self.window.core.config.get('llama.idx.mode'),
                    external_functions=functions,
                    tools_outputs=tools_outputs,
                    max_tokens=max_tokens,
                    is_expert_call=True,  # mark as expert call
                    preset=expert,
                )
                extra = {}
                if use_index:
                    extra["agent_idx"] = db_idx

                tools = self.window.core.agents.tools.prepare(
                    bridge_context, extra, verbose=False, force=True)

                # remove expert_call tool from tools
                for tool in list(tools):
                    if tool.metadata.name == TOOL_EXPERT_CALL_NAME:
                        tools.remove(tool)

                result = self.call_agent(
                    context=bridge_context,
                    tools=tools,
                    ctx=ctx,
                    query=query,
                    llm=llm,
                    system_prompt=sys_prompt,
                    verbose=verbose,
                )
                ctx.reply = False  # reset reply flag, we handle reply here

                if not result:  # abort if bridge call failed
                    self.signals.finished.emit()
                    return
            else:
                # native func call
                if self.window.core.command.is_native_enabled(force=False, model=model):

                    # get native functions, without expert_call here
                    functions = self.window.core.command.get_functions(master_ctx.id)

                    # append retrieval tool if index is selected
                    if use_index:
                        retriever_tool = self.window.core.experts.get_retriever_tool()
                        func_list = self.window.core.command.cmds_to_functions([retriever_tool])
                        functions.append(func_list[0])  # append only first function

                # call bridge
                bridge_context = BridgeContext(
                    ctx=ctx,
                    history=history,
                    mode=mode,
                    parent_mode=base_mode,
                    model=model_data,
                    system_prompt=sys_prompt,
                    system_prompt_raw=sys_prompt_raw,
                    prompt=query,
                    stream=False,
                    attachments=files,
                    file_ids=file_ids,
                    assistant_id=self.window.core.config.get('assistant'),
                    idx=db_idx,
                    idx_mode=self.window.core.config.get('llama.idx.mode'),
                    external_functions=functions,
                    tools_outputs=tools_outputs,
                    max_tokens=max_tokens,
                    is_expert_call=True,  # mark as expert call
                    preset=expert,
                    force_sync=True,  # force sync call, no async bridge call
                    request=True,  # use normal request instead of quick call
                )

                self.signals.lock_input.emit()  # emit lock input signal
                event = KernelEvent(KernelEvent.CALL, {
                    'context': bridge_context,  # call using slave ctx history
                    'extra': {},
                })
                self.window.dispatch(event)
                result = event.data.get("response")
                # result: <tool>{"cmd": "read_file", "params": {"path": ["xxxx.txt"]}}</tool>
                # ctx:
                # input: please read the file xxx.txt
                # output: <tool>cmd read</tool>
                if not result and not ctx.tool_calls:  # abort if bridge call failed
                    self.signals.finished.emit()
                    return

            # handle output
            ctx.current = False  # reset current state
            ctx.output = result  # store expert output in their context

            self.window.core.ctx.update_item(ctx)

            ctx.from_previous()  # append previous result if exists
            ctx.clear_reply()  # reset results

            if not use_agent:
                ctx.sub_tool_call = True
                self.signals.cmd.emit(ctx, master_ctx, expert_id, expert_name, result)  # emit cmd signal
                # tool call here and reply to window, from <tool></tool>
                return

            # if command to execute then end here, and reply is returned to reply() above from stack, and ctx.reply = TRUE here
            ctx.from_previous()  # append previous result again before save
            self.window.core.ctx.update_item(ctx)  # update ctx in DB

            # if commands reply after bridge call, then stop (already handled in sync dispatcher)
            if ctx.reply:
                self.signals.finished.emit()
                return

            # make copy of ctx for reply, and change input name to expert name
            reply_ctx = CtxItem()
            reply_ctx.from_dict(ctx.to_dict())
            reply_ctx.meta = master_ctx.meta

            # assign expert output
            reply_ctx.output = result
            reply_ctx.input_name = expert_name
            reply_ctx.output_name = ""
            reply_ctx.cmds = []  # clear cmds
            reply_ctx.sub_call = True  # this flag is not copied in to_dict

            # reply to main thread

            # send to reply()
            # input: something (no tool results here)
            # output: ... (call the master)
            self.signals.response.emit(reply_ctx, str(expert_id))  # emit response signal

        except Exception as e:
            self.window.core.debug.log(e)
            self.signals.error.emit(str(e))

        finally:
            self.signals.finished.emit()
            self.cleanup()

    def call_agent(
            self,
            context: BridgeContext,
            tools: Optional[List[QueryEngineTool]] = None,
            ctx: Optional[CtxItem] = None,
            query: str = "",
            llm=None,
            system_prompt: str = "",
            verbose: bool = False,

    ) -> str:
        """
        Call agent with tools and index

        :param context: Bridge context
        :param tools: Tools
        :param ctx: CtxItem
        :param query: Input prompt
        :param llm: LLM provider
        :param system_prompt: System prompt to use for agent
        :param verbose: Verbose mode, default is False
        :return: Response from agent as string
        """
        history = self.window.core.agents.memory.prepare(context)
        bridge_context = BridgeContext(
            ctx=ctx,
            system_prompt=system_prompt,
            model=context.model,
            prompt=query,
            stream=False,
            is_expert_call=True,  # mark as expert call
        )
        extra = {
            "agent_provider": "react",  # use react workflow provider
            "agent_idx": context.idx,  # index to use
            "agent_tools": tools,  # tools to use
            "agent_history": history,  # already prepared history
        }
        response_ctx = self.window.core.agents.runner.call_once(
            context=bridge_context,
            extra=extra,
            signals=None,
        )
        if response_ctx:
            return str(response_ctx.output)
        else:
            return "No response from expert."

    def cleanup(self):
        """Cleanup resources after worker execution."""
        sig = self.signals
        self.signals = None
        if sig is not None:
            try:
                sig.deleteLater()
            except RuntimeError:
                pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.28 00:00:00                  #
# ================================================== #

import threading
from typing import Any, Dict, Optional, Union, List

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QApplication

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AGENT_OPENAI,
    MODE_ASSISTANT,
    MODE_EXPERT,
    MODE_LLAMA_INDEX,
)
from pygpt_net.core.events import KernelEvent, RenderEvent, BaseEvent, RealtimeEvent, Event
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .reply import Reply
from .stack import Stack


class Kernel:

    STATE_IDLE = "idle"
    STATE_BUSY = "busy"
    STATE_ERROR = "error"

    __slots__ = (
        "window",
        "replies",
        "stack",
        "halt",
        "busy",
        "last_stack",
        "status",
        "state",
        "not_stop_on_events",
    )

    _INPUT_EVENTS = frozenset((KernelEvent.INPUT_USER, KernelEvent.INPUT_SYSTEM))
    _REPLY_EVENTS = frozenset((KernelEvent.REPLY_ADD, KernelEvent.REPLY_RETURN))
    _STATE_EVENTS = frozenset((KernelEvent.STATE_BUSY, KernelEvent.STATE_IDLE, KernelEvent.STATE_ERROR))

    _REQUEST_EVENTS = frozenset((KernelEvent.REQUEST, KernelEvent.REQUEST_NEXT))
    _OUTPUT_EVENTS = frozenset(
        (
            KernelEvent.RESPONSE_OK,
            KernelEvent.RESPONSE_ERROR,
            KernelEvent.RESPONSE_FAILED,
            KernelEvent.APPEND_BEGIN,
            KernelEvent.APPEND_DATA,
            KernelEvent.APPEND_END,
            KernelEvent.LIVE_APPEND,
            KernelEvent.LIVE_CLEAR,
        )
    )
    _STACK_ADD_EVENTS = frozenset((KernelEvent.TOOL_CALL, KernelEvent.AGENT_CONTINUE, KernelEvent.AGENT_CALL))
    _CALL_EVENTS = frozenset((KernelEvent.CALL, KernelEvent.FORCE_CALL))
    _QUEUE_EVENTS_ALL = _REQUEST_EVENTS | _OUTPUT_EVENTS | _STACK_ADD_EVENTS | _CALL_EVENTS

    _ASYNC_DISABLED_MODES = frozenset(
        (MODE_ASSISTANT, MODE_AGENT, MODE_EXPERT, MODE_AGENT_LLAMA, MODE_AGENT_OPENAI, MODE_LLAMA_INDEX)
    )
    _THREADED_MODES = frozenset((MODE_AGENT_LLAMA, MODE_AGENT_OPENAI))

    def __init__(self, window=None):
        """
        Initialize the Kernel with a window context.

        :param window: The window context to which the kernel will be bound.
        """
        self.window = window
        self.replies = Reply(window)
        self.stack = Stack(window)
        self.halt = False
        self.busy = False
        self.last_stack = []
        self.status = ""
        self.state = self.STATE_IDLE
        self.not_stop_on_events = [
            KernelEvent.APPEND_DATA,
            KernelEvent.INPUT_USER,
            KernelEvent.FORCE_CALL,
            KernelEvent.STATUS,
            Event.AUDIO_INPUT_RECORD_TOGGLE,
            RealtimeEvent.RT_INPUT_AUDIO_DELTA,
            RealtimeEvent.RT_INPUT_AUDIO_MANUAL_STOP,
            RealtimeEvent.RT_INPUT_AUDIO_MANUAL_START,
            RealtimeEvent.RT_OUTPUT_AUDIO_COMMIT,
            RealtimeEvent.RT_OUTPUT_TURN_END,
            RealtimeEvent.RT_OUTPUT_READY,
        ]

    def init(self):
        """
        Initialize the kernel state.
        """
        self.last_stack = []
        self.halt = False
        self.busy = False
        self.state = self.STATE_IDLE
        self.status = ""

    @Slot(object)
    def listener(self, event: BaseEvent):
        """
        Handle incoming events from the window's event dispatcher.

        :param event: BaseEvent: The event to handle.
        """
        if self.halt and event.name not in self.not_stop_on_events:
            return
        self.window.dispatch(event)

    def handle(self, event: KernelEvent):
        """
        Handle the incoming kernel event and process it based on its type.

        :param event: KernelEvent: The event to handle.
        """
        if self.halt and event.name not in self.not_stop_on_events:
            return

        self.store(event)
        name = event.name
        data = event.data
        context = data.get("context")
        extra = data.get("extra")
        response = data.get("response")

        if name in self._INPUT_EVENTS:
            response = self.input(context, extra, event)
        elif name in self._QUEUE_EVENTS_ALL:
            response = self.queue(context, extra, event)
        elif name in self._REPLY_EVENTS:
            response = self.reply(context, extra, event)
        elif name in self._STATE_EVENTS:
            self.set_state(event)
        elif name == KernelEvent.STATUS:
            self.set_status(data.get("status"))

        data["response"] = response

    def input(
        self,
        context: BridgeContext,
        extra: Dict[str, Any],
        event: KernelEvent,
    ):
        """
        Handle input events, such as user input or system input.

        :param context: BridgeContext: The context of the input event.
        :param extra: Dict[str, Any]: Additional data related to the input event.
        :param event: KernelEvent: The event that triggered the input handling.
        :return: Optional[Union[bool, str]]:
        """
        if self.halt and event.name != KernelEvent.INPUT_USER:
            return
        if event.name in self._INPUT_EVENTS:
            w = self.window
            return w.controller.chat.input.send(context, extra)

    def queue(
        self,
        context: BridgeContext,
        extra: Dict[str, Any],
        event: KernelEvent,
    ) -> Optional[Union[bool, str, List[Dict]]]:
        if self.halt and event.name not in self.not_stop_on_events:
            return
        name = event.name
        if name in self._REQUEST_EVENTS:
            return self.call(context, extra, event)
        elif name in self._OUTPUT_EVENTS:
            return self.output(context, extra, event)
        elif name in self._STACK_ADD_EVENTS:
            return self.stack.add(context.reply_context)
        elif name in self._CALL_EVENTS:
            return self.call(context, extra, event)

    def call(
        self,
        context: BridgeContext,
        extra: Dict[str, Any],
        event: KernelEvent,
    ) -> Optional[Union[bool, str]]:
        """
        Handle call events, which may involve making requests or executing commands.

        :param context: BridgeContext: The context of the call event.
        :param extra: Dict[str, Any]: Additional data related to the call event.
        :param event: KernelEvent: The event that triggered the call handling.
        :return: Optional[Union[bool, str]]:
        """
        if self.halt and event.name not in self.not_stop_on_events:
            return
        w = self.window
        name = event.name
        if name == KernelEvent.REQUEST:
            return w.core.bridge.request(context, extra)
        elif name == KernelEvent.REQUEST_NEXT:
            return w.core.bridge.request_next(context, extra)
        elif name in self._CALL_EVENTS:
            return w.core.bridge.call(context, extra)

    def reply(
        self,
        context: BridgeContext,
        extra: Dict[str, Any],
        event: KernelEvent,
    ) -> Optional[List[Dict]]:
        """
        Handle reply events, which may involve adding replies or returning input.

        :param context: BridgeContext: The context of the reply event.
        :param extra: Dict[str, Any]: Additional data related to the reply event.
        :param event: KernelEvent: The event that triggered the reply handling.
        :return: Optional[List[Dict]]:
        """
        if self.halt:
            return
        if event.name == KernelEvent.REPLY_ADD:
            return self.replies.add(context, extra)
        elif event.name == KernelEvent.REPLY_RETURN:
            return self.input(context, extra, KernelEvent(KernelEvent.INPUT_SYSTEM))

    def output(
        self,
        context: BridgeContext,
        extra: Dict[str, Any],
        event: KernelEvent,
    ):
        """
        Handle output events, which may involve processing responses or appending data.

        :param context: BridgeContext: The context of the output event.
        :param extra: Dict[str, Any]: Additional data related to the output event.
        :param event: KernelEvent: The event that triggered the output handling.
        :return: Optional[Union[bool, str]]:
        """
        if self.halt and event.name not in self.not_stop_on_events:
            return
        w = self.window
        resp = w.controller.chat.response
        name = event.name
        if name == KernelEvent.RESPONSE_OK:
            return resp.handle(context, extra, True)
        elif name == KernelEvent.RESPONSE_ERROR:
            return resp.handle(context, extra, False)
        elif name == KernelEvent.RESPONSE_FAILED:
            return resp.failed(context, extra)
        elif name == KernelEvent.APPEND_BEGIN:
            return resp.begin(context, extra)
        elif name == KernelEvent.APPEND_DATA:
            return resp.append(context, extra)
        elif name == KernelEvent.APPEND_END:
            return resp.end(context, extra)
        elif name == KernelEvent.LIVE_APPEND:
            return resp.live_append(context, extra)
        elif name == KernelEvent.LIVE_CLEAR:
            return resp.live_clear(context, extra)

    def restart(self):
        """
        Restart the kernel by dispatching a restart event and re-initializing the state.
        """
        self.window.dispatch(KernelEvent(KernelEvent.RESTART))
        self.init()

    def terminate(self):
        """
        Terminate the kernel by dispatching a terminate event, stopping the window, and destroying plugins.
        """
        self.window.dispatch(KernelEvent(KernelEvent.TERMINATE))
        self.stop(exit=True)
        self.window.controller.plugins.destroy()
        self.window.controller.realtime.shutdown()

    def stop(self, exit: bool = False):
        """
        Stop the kernel and its associated processes.

        :param exit: If True, exit the application after stopping.
        """
        self.halt = True
        w = self.window
        w.controller.chat.common.stop(exit=exit)
        w.controller.audio.stop_audio()
        if not exit:
            w.dispatch(KernelEvent(KernelEvent.STOP))
            self.set_state(KernelEvent(KernelEvent.STATE_IDLE, {"msg": trans("status.stopped")}))

    def set_state(self, event: KernelEvent):
        """
        Set the kernel state based on the event received.

        :param event: KernelEvent: The event containing the state information.
        """
        name = event.name
        w = self.window
        tray = w.ui.tray
        is_main = self.is_main_thread()

        if name == KernelEvent.STATE_BUSY:
            self.busy = True
            self.state = self.STATE_BUSY
            tray.set_icon(self.STATE_BUSY)
            if not self.halt and is_main:
                w.dispatch(RenderEvent(RenderEvent.STATE_BUSY))
        elif name == KernelEvent.STATE_IDLE:
            self.busy = False
            self.state = self.STATE_IDLE
            tray.set_icon(self.STATE_IDLE)
            if is_main:
                w.dispatch(RenderEvent(RenderEvent.STATE_IDLE))
        elif name == KernelEvent.STATE_ERROR:
            self.busy = False
            self.state = self.STATE_ERROR
            tray.set_icon(self.STATE_ERROR)
            if is_main:
                w.dispatch(RenderEvent(RenderEvent.STATE_ERROR))

        msg = event.data.get("msg", None)
        if msg is not None:
            self.set_status(msg)

    def set_status(self, status: str):
        """
        Set the status message for the kernel and update the UI.

        :param status: str: The status message to set.
        """
        self.status = status
        self.window.ui.status(status)
        QApplication.processEvents()  # process events to update UI

    def resume(self):
        """
        Resume the kernel operation after it has been halted.
        """
        self.halt = False

    def stopped(self) -> bool:
        """
        Check if the kernel is currently halted.

        :return: bool: True if the kernel is halted, False otherwise.
        """
        return self.halt

    def store(self, event):
        """
        Store the event in the kernel's stack for later processing.

        :param event: BaseEvent: The event to store.
        """
        return

    def async_allowed(self, ctx: CtxItem) -> bool:
        """
        Check if asynchronous operations are allowed based on the current mode and context.

        :param ctx: CtxItem: The context item containing information about the current operation.
        :return: bool: True if asynchronous operations are allowed, False otherwise.
        """
        if self.window.core.config.get("mode") in self._ASYNC_DISABLED_MODES:
            return False
        if ctx.agent_call:
            return False
        controller = self.window.controller
        agent = controller.agent
        if agent.legacy.enabled() or agent.experts.enabled():
            return False
        return True

    def is_threaded(self) -> bool:
        """
        Check if the current mode allows for threaded operations.

        :return: bool: True if the current mode is threaded, False otherwise.
        """
        return self.window.core.config.get("mode") in self._THREADED_MODES

    def is_main_thread(self) -> bool:
        """
        Check if the current thread is the main thread.

        :return: bool: True if the current thread is the main thread, False otherwise.
        """
        return threading.current_thread() is threading.main_thread()

    def close_clients(self):
        """
        Close all active clients associated with the kernel.
        """
        w = self.window
        try:
            w.core.api.openai.safe_close()
            w.core.api.google.safe_close()
        except Exception:
            pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.25 18:00:00                  #
# ================================================== #

import os
import sys
import shutil
import subprocess

from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices

# QtDBus is optional and may not be available on all systems
try:
    from PySide6.QtDBus import QDBusInterface, QDBusConnection, QDBusMessage
    HAS_QT_DBUS = True
except Exception:
    HAS_QT_DBUS = False

IS_WINDOWS = sys.platform.startswith("win")
IS_MAC = sys.platform == "darwin"
IS_LINUX = sys.platform.startswith("linux")

class Opener:
    # ===== public API =====
    @staticmethod
    def open_path(path: str, reveal: bool = False) -> bool:
        """
        Open file or directory in the system file manager or default application.

        :param path: Path to file or directory
        :param reveal: If True and path is a file, reveal it in the file manager
        :return: True if successful, False otherwise
        """
        p = os.path.abspath(path)

        if IS_WINDOWS:
            if reveal and os.path.isfile(p):
                return Opener._reveal_windows(p)
            if os.path.isdir(p):
                return Opener._open_dir_windows(p)
            return Opener._open_file_windows(p)

        if IS_MAC:
            if reveal and os.path.isfile(p):
                return Opener._reveal_mac(p)
            if os.path.isdir(p):
                return Opener._open_dir_mac(p)
            return Opener._open_file_mac(p)

        # Linux
        if reveal and os.path.isfile(p):
            if Opener._reveal_linux(p):
                return True
            # fallback:
            return Opener._open_dir_linux(os.path.dirname(p) or "/")
        if os.path.isdir(p):
            return Opener._open_dir_linux(p)
        return Opener._open_file_linux(p)

    # ===== Windows =====
    @staticmethod
    def _open_dir_windows(path: str) -> bool:
        """
        Open directory in Windows Explorer

        :param path: Path to directory
        :return: True if successful, False otherwise
        """
        try:
            os.startfile(path)
            return True
        except Exception:
            try:
                subprocess.Popen(["explorer", path])
                return True
            except Exception:
                return False

    @staticmethod
    def _reveal_windows(path: str) -> bool:
        """
        Reveal file in Windows Explorer

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        try:
            subprocess.Popen(["explorer", "/select,", os.path.normpath(path)])
            return True
        except Exception:
            return Opener._open_dir_windows(os.path.dirname(path) or "C:\\")

    @staticmethod
    def _open_file_windows(path: str) -> bool:
        """
        Open file with default application

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        try:
            os.startfile(path)
            return True
        except Exception:
            # Qt
            return QDesktopServices.openUrl(QUrl.fromLocalFile(path))

    # ===== macOS =====
    @staticmethod
    def _open_dir_mac(path: str) -> bool:
        """
        Open directory in Finder

        :param path: Path to directory
        :return: True if successful, False otherwise
        """
        try:
            subprocess.Popen(["open", path])
            return True
        except Exception:
            return QDesktopServices.openUrl(QUrl.fromLocalFile(path))

    @staticmethod
    def _reveal_mac(path: str) -> bool:
        """
        Reveal file in Finder

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        try:
            subprocess.Popen(["open", "-R", path])
            return True
        except Exception:
            return Opener._open_dir_mac(os.path.dirname(path) or "/")

    @staticmethod
    def _open_file_mac(path: str) -> bool:
        """
        Open file with default application

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        try:
            subprocess.Popen(["open", path])
            return True
        except Exception:
            return QDesktopServices.openUrl(QUrl.fromLocalFile(path))

    # ===== Linux =====
    @staticmethod
    def _fm_iface():
        """
        Get FileManager1 D-Bus interface if available

        :return: QDBusInterface or None
        """
        if not (IS_LINUX and HAS_QT_DBUS):
            return None
        bus = QDBusConnection.sessionBus()
        iface = QDBusInterface(
            "org.freedesktop.FileManager1",
            "/org/freedesktop/FileManager1",
            "org.freedesktop.FileManager1",
            bus,
        )
        return iface if iface.isValid() else None

    @staticmethod
    def _show_folders_dbus(paths):
        """
        Show folders using D-Bus FileManager1 interface

        :param paths: List of folder paths
        :return: True if successful, False otherwise
        """
        iface = Opener._fm_iface()
        if not iface:
            return False
        uris = [QUrl.fromLocalFile(p).toString() for p in paths]
        reply = iface.call("ShowFolders", uris, "")
        return reply.type() != QDBusMessage.ErrorMessage

    @staticmethod
    def _show_items_dbus(paths):
        """
        Show items using D-Bus FileManager1 interface

        :param paths: List of item paths
        :return: True if successful, False otherwise
        """
        iface = Opener._fm_iface()
        if not iface:
            return False
        uris = [QUrl.fromLocalFile(p).toString() for p in paths]
        reply = iface.call("ShowItems", uris, "")
        return reply.type() != QDBusMessage.ErrorMessage

    @staticmethod
    def _open_with_cli_linux(path):
        """
        Open path using common CLI tools (xdg-open, gio)

        :param path: Path to file or directory
        :return: True if successful, False otherwise
        """
        for cmd in (["xdg-open", path], ["gio", "open", path]):
            if shutil.which(cmd[0]):
                try:
                    subprocess.Popen(cmd)
                    return True
                except Exception:
                    pass
        return False

    @staticmethod
    def _open_dir_linux(path: str) -> bool:
        """
        Open directory in default file manager

        :param path: Path to directory
        :return: True if successful, False otherwise
        """
        if Opener._show_folders_dbus([path]):
            return True
        if QDesktopServices.openUrl(QUrl.fromLocalFile(path)):
            return True
        return Opener._open_with_cli_linux(path)

    @staticmethod
    def _reveal_linux(path: str) -> bool:
        """
        Reveal file in default file manager

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        if Opener._show_items_dbus([path]):
            return True
        return False

    @staticmethod
    def _open_file_linux(path: str) -> bool:
        """
        Open file with default application

        :param path: Path to file
        :return: True if successful, False otherwise
        """
        if QDesktopServices.openUrl(QUrl.fromLocalFile(path)):
            return True
        return Opener._open_with_cli_linux(path)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.28 08:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QSplitter, QWidget, QSizePolicy, \
    QTabWidget, QFileDialog, QScrollArea, QFrame

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
    MODE_COMPUTER,
    MODE_AGENT_OPENAI,
)
from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.editor import EditorDialog
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.lists.experts import ExpertsEditor
from pygpt_net.utils import trans


class Preset(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Preset, self).__init__(window, *args, **kwargs)
        """
        Preset editor dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "preset"
        self.dialog_id = "preset.presets"

    def setup(self):
        """Setup preset editor dialog"""
        self.window.ui.nodes['preset.btn.current'] = QPushButton(trans("dialog.preset.btn.current"))
        self.window.ui.nodes['preset.btn.save'] = QPushButton(trans("dialog.preset.btn.save"))
        self.window.ui.nodes['preset.btn.current'].clicked.connect(self.window.controller.presets.editor.from_current)
        self.window.ui.nodes['preset.btn.save'].clicked.connect(self.window.controller.presets.editor.save)

        self.window.ui.nodes['preset.btn.current'].setAutoDefault(False)
        self.window.ui.nodes['preset.btn.save'].setAutoDefault(True)

        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['preset.btn.current'])
        footer.addWidget(self.window.ui.nodes['preset.btn.save'])

        # fields
        self.window.ui.paths[self.id] = QLabel(str(self.window.core.config.path))

        # get option fields config
        fields = self.window.controller.presets.editor.get_options()

        # build settings widgets
        widgets = self.build_widgets(self.id, fields, excluded=["tool.function"])  # from base config dialog

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[self.id][key] = widgets[key]

        # apply widgets to layouts
        options = {}
        for key in widgets:
            if key == "tool.function":
                continue
            if fields[key]["type"] in ['text', 'int', 'float']:
                # if key != "prompt":  # built separately
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'textarea':
                if key == "prompt":
                    widgets[key].setMinimumHeight(100)
                    continue
                if key == "description":
                    widgets[key].setMinimumHeight(50)
                else:
                    widgets[key].setMinimumHeight(100)
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool':
                options[key] = self.add_raw_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'dict':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'combo':
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool_list':
                options[key] = self.add_row_option(widgets[key], fields[key])

        rows = QVBoxLayout()

        for key in options:
            options[key].setContentsMargins(0, 0, 0, 0)

        # modes
        mode_keys_left = [
            MODE_CHAT,
            MODE_LLAMA_INDEX,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        mode_keys_middle = [
            MODE_COMPLETION,
            MODE_IMAGE,
            # MODE_VISION,
            MODE_COMPUTER,
            MODE_EXPERT,
        ]
        mode_keys_right = [
            MODE_AGENT_LLAMA,
            MODE_AGENT_OPENAI,
            MODE_AGENT,
        ]

        rows_mode_left = QVBoxLayout()
        rows_mode_left.setContentsMargins(0, 0, 0, 0)
        for key in mode_keys_left:
            rows_mode_left.addLayout(options[key])
        rows_mode_left.addStretch()

        rows_mode_middle = QVBoxLayout()
        rows_mode_middle.setContentsMargins(0, 0, 0, 0)
        for key in mode_keys_middle:
            rows_mode_middle.addLayout(options[key])
        rows_mode_middle.addStretch()

        rows_mode_right = QVBoxLayout()
        rows_mode_right.setContentsMargins(0, 0, 0, 0)
        for key in mode_keys_right:
            rows_mode_right.addLayout(options[key])
        rows_mode_right.addStretch()

        rows_mode = QHBoxLayout()
        rows_mode.addLayout(rows_mode_left)
        rows_mode.addLayout(rows_mode_middle)
        rows_mode.addLayout(rows_mode_right)
        rows_mode.setAlignment(Qt.AlignTop)
        rows_mode.setContentsMargins(0, 0, 0, 0)
        rows_mode.addStretch(1)

        # modes
        modes = QWidget()
        modes.setLayout(rows_mode)
        modes.setContentsMargins(0, 0, 0, 0)
        # Ensure modes column never grows taller than its content; avoids creating vertical gap above the splitter.
        modes.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        self.window.ui.nodes['preset.editor.modes'] = modes

        # experts
        self.window.ui.nodes['preset.editor.experts'] = ExpertsEditor(self.window)

        # desc and prompt
        desc = QWidget()
        desc.setLayout(options['description'])
        desc.setContentsMargins(0, 5, 0, 5)
        self.window.ui.nodes['preset.editor.description'] = desc

        # prompt + extra options
        prompt_layout = QVBoxLayout()
        prompt_layout.addWidget(widgets['prompt'])
        prompt_layout.setContentsMargins(0, 10, 0, 10)

        footer_layout = self.prepare_extra_config(prompt_layout)

        prompt_layout = QVBoxLayout()
        prompt_layout.setContentsMargins(0, 0, 0, 0)
        prompt_layout.addWidget(self.window.ui.nodes['preset.editor.description'])
        prompt_layout.addLayout(footer_layout)

        widget_prompt = QWidget()
        widget_prompt.setLayout(prompt_layout)
        widget_prompt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        widget_prompt.setMinimumHeight(300)

        # left column
        left_keys = [
            "filename",
            "name",
            "model",
            "temperature",
            "agent_provider",
            "agent_provider_openai",
            "idx",
        ]
        # personalize tab
        personalize_keys = [
            "ai_name",
            "user_name",
            "ai_avatar",
            "ai_personalize",
        ]
        for key in left_keys:
            option_layout = options[key]
            if key in ["agent_provider", "agent_provider_openai"]:
                # special case with settings icon on right
                node_layout = QHBoxLayout()
                builder_btn = QPushButton(QIcon(":/icons/robot.svg"), "")
                builder_btn.setToolTip(trans("agent.builder.tooltip"))
                builder_btn.clicked.connect(lambda: self.window.tools.get("agent_builder").toggle())
                node_layout.setContentsMargins(0, 0, 0, 0)
                node_layout.addLayout(options[key])
                node_layout.addWidget(builder_btn)
                # align top right
                node_layout.setAlignment(builder_btn, Qt.AlignTop | Qt.AlignRight)
                option_layout = node_layout

            node_key = f"preset.editor.{key}"
            node = QWidget()
            node.setLayout(option_layout)
            node.setContentsMargins(0, 0, 0, 0)
            rows.addWidget(node)
            self.window.ui.nodes[node_key] = node

        # remote tools
        remote_tools =  QWidget()
        remote_tools.setLayout(options['remote_tools'])
        remote_tools.setContentsMargins(0, 0, 0, 0)
        self.window.ui.nodes['preset.editor.remote_tools'] = remote_tools

        rows_remote_tools = QVBoxLayout()
        rows_remote_tools.addWidget(self.window.ui.nodes['preset.editor.remote_tools'])
        rows_remote_tools.addStretch(1)

        widget_remote_tools = QWidget()
        widget_remote_tools.setLayout(rows_remote_tools)

        # personalize
        personalize_rows = QVBoxLayout()
        for key in personalize_keys:
            node_key = f"preset.editor.{key}"
            node = QWidget()
            node.setLayout(options[key])
            node.setContentsMargins(0, 0, 0, 0)
            personalize_rows.addWidget(node)
            self.window.ui.nodes[node_key] = node

        self.window.ui.nodes['preset.editor.ai_avatar'].setVisible(False)

        # warning label
        warn_label = HelpLabel(trans("preset.personalize.warning"))

        # avatar
        self.window.ui.nodes['preset.editor.avatar'] = AvatarWidget(self.window)
        personalize_rows.addWidget(self.window.ui.nodes['preset.editor.avatar'])
        personalize_rows.addStretch(1)
        personalize_rows.addWidget(warn_label)
        personalize_rows.setContentsMargins(10, 30, 10, 0)

        rows.setContentsMargins(0, 0, 0, 0)
        rows.addStretch(1)
        rows.setAlignment(Qt.AlignTop)

        widget_base = QWidget()
        widget_base.setLayout(rows)
        widget_base.setMinimumWidth(300)
        # Keep base column at content height; do not stretch vertically.
        widget_base.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)

        self.window.ui.nodes['preset.editor.experts'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main = QHBoxLayout()
        main.addWidget(widget_base)
        main.addWidget(self.window.ui.nodes['preset.editor.modes'])

        widget_main = QWidget()
        widget_main.setLayout(main)
        # Critical: ensure the whole top pane (base + modes) stays at content height.
        widget_main.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        splitter = QSplitter(Qt.Vertical)
        splitter.addWidget(widget_main)
        splitter.addWidget(widget_prompt)
        splitter.setChildrenCollapsible(False)  # avoid accidental collapsing of any pane
        # All extra vertical space goes to bottom (extra agent options); top stays at its sizeHint.
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)
        self.window.ui.splitters['editor.presets'] = splitter

        widget_personalize = QWidget()
        widget_personalize.setLayout(personalize_rows)

        experts_rows = QVBoxLayout()
        experts_rows.addWidget(self.window.ui.nodes['preset.editor.experts'])

        widget_experts = QWidget()
        widget_experts.setLayout(experts_rows)

        tabs = QTabWidget()
        tabs.addTab(splitter, trans("preset.tab.general"))
        tabs.addTab(widget_personalize, trans("preset.tab.personalize"))
        tabs.addTab(widget_experts, trans("preset.tab.experts"))
        tabs.addTab(widget_remote_tools, trans("preset.tab.remote_tools"))
        self.window.ui.tabs['preset.editor.tabs'] = tabs

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.tabs['preset.editor.tabs'])
        layout.addLayout(footer)

        dialog = EditorDialog(self.window, self.dialog_id)
        dialog.setSizeGripEnabled(True)
        dialog.setWindowFlags(
            dialog.windowFlags() | Qt.WindowMaximizeButtonHint
        )
        dialog.setLayout(layout)
        dialog.setWindowTitle(trans('dialog.preset'))
        dialog.on_close_callback = self.on_close
        self.window.ui.dialog['editor.' + self.dialog_id] = dialog


    def prepare_extra_config(self, prompt_layout):
        """
        Build extra configuration for the preset editor dialog

        :param prompt_layout: Layout for the prompt editor
        """
        prompt_layout.setContentsMargins(0, 10, 0, 10)
        prompt_widget = QWidget()
        prompt_widget.setLayout(prompt_layout)

        self.window.ui.nodes['preset.editor.extra'] = {}

        tabs = QTabWidget()

        # Make the prompt tab scrollable to avoid vertical overlap in narrow layouts.
        scroll_prompt = QScrollArea()
        scroll_prompt.setWidget(prompt_widget)
        scroll_prompt.setWidgetResizable(True)
        scroll_prompt.setFrameShape(QFrame.NoFrame)
        tabs.addTab(
            scroll_prompt,
            trans("preset.prompt"),
        )
        tabs.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        tabs.setMinimumHeight(150)
        layout = QVBoxLayout()
        layout.addWidget(tabs)
        self.window.ui.tabs['preset.editor.extra'] = tabs
        return layout


    def build_option_widgets(self, id: str, fields: dict) -> tuple:
        """
        Build option widgets for the preset editor dialog

        :param id: Agent config unique id
        :param fields: Dictionary of options
        :return: tuple of widgets and options (layouts)
        """
        if id not in self.window.ui.config:
            self.window.ui.config[id] = {}

        widgets = self.build_widgets(id, fields)  # from base config dialog

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[id][key] = widgets[key]

        # apply widgets to layouts
        options = {}
        for key in widgets:
            if fields[key]["type"] in ['text', 'int', 'float']:
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'textarea':
                widgets[key].setMinimumHeight(100)
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool':
                widgets[key].setMaximumHeight(38)
                options[key] = self.add_raw_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'dict':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'combo':
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool_list':
                options[key] = self.add_row_option(widgets[key], fields[key])

        return widgets, options

    # on close
    def on_close(self):
        """Close event callback"""
        self.window.controller.presets.select_current(no_scroll=True)
        self.window.controller.presets.editor.opened = False


class AvatarWidget(QWidget):
    def __init__(self, window=None):
        super().__init__(window)
        self.window = window
        self.avatar_preview = None
        self.choose_button = None
        self.remove_button = None
        self.path_line_edit = None
        self.init_ui()

    def init_ui(self):
        """Initialize the avatar widget UI."""
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)

        current_avatar_label = QLabel(trans("preset.personalize.avatar.current"))
        # Center the section title for better visual balance.
        current_avatar_label.setAlignment(Qt.AlignHCenter)
        main_layout.addWidget(current_avatar_label)

        self.avatar_preview = QLabel(self)
        self.avatar_preview.setFixedSize(200, 200)
        # Keep a visible border, but round it to visually indicate the circular avatar.
        self.avatar_preview.setStyleSheet("border: 1px solid gray; border-radius: 100px; background: transparent;")
        self.avatar_preview.setAlignment(Qt.AlignCenter)
        # Center the avatar widget within the column.
        main_layout.addWidget(self.avatar_preview, 0, Qt.AlignHCenter)

        buttons_layout = QHBoxLayout()

        self.choose_button = QPushButton(trans("preset.personalize.avatar.choose"), self)
        self.choose_button.clicked.connect(self.open_file_dialog)
        buttons_layout.addWidget(self.choose_button)

        self.remove_button = QPushButton(trans("preset.personalize.avatar.remove"), self)
        self.remove_button.clicked.connect(self.window.controller.presets.editor.remove_avatar)
        self.remove_button.setEnabled(False)
        buttons_layout.addWidget(self.remove_button)
        buttons_layout.setContentsMargins(0, 10, 0, 0)
        # Center the action buttons below the avatar.
        buttons_layout.setAlignment(Qt.AlignCenter)

        main_layout.addLayout(buttons_layout)
        main_layout.setContentsMargins(0, 10, 0, 0)
        main_layout.addStretch()

    def open_file_dialog(self):
        """Open a file dialog to select an avatar image file."""
        file_name, _ = QFileDialog.getOpenFileName(
            self, trans("preset.personalize.avatar.choose.title"), "", "Images (*.png *.jpg *.jpeg *.bmp *.gif *.webp)"
        )
        if file_name:
            self.window.controller.presets.editor.upload_avatar(file_name)

    def load_avatar(self, file_path: str):
        """
        Load and display the avatar image from the given file path.

        :param file_path: Path to the avatar image file
        """
        from PySide6.QtGui import QPixmap
        pixmap = QPixmap(file_path)
        if not pixmap.isNull():
            cover_pix = self.get_cover_pixmap(pixmap, self.avatar_preview.width(), self.avatar_preview.height())
            # Render the avatar as a circular pixmap.
            circle_pix = self.get_circular_pixmap(cover_pix, self.avatar_preview.width(), self.avatar_preview.height())
            self.avatar_preview.setPixmap(circle_pix)
        self.remove_button.setEnabled(True)

    def enable_remove_button(self, enabled: bool = True):
        """
        Enable or disable the remove button based on the presence of an avatar.

        :param enabled: True to enable, False to disable
        """
        self.remove_button.setEnabled(enabled)

    def disable_remove_button(self):
        """Disable the remove button."""
        self.enable_remove_button(False)

    def get_cover_pixmap(self, pixmap, target_width, target_height):
        """
        Scale and crop the pixmap to fit the target dimensions while maintaining aspect ratio.

        :param pixmap: Original pixmap
        :param target_width: Target width for the avatar preview
        :param target_height: Target height for the avatar preview
        :return: Scaled and cropped pixmap
        """
        factor = max(target_width / pixmap.width(), target_height / pixmap.height())
        new_width = int(pixmap.width() * factor)
        new_height = int(pixmap.height() * factor)
        scaled_pix = pixmap.scaled(new_width, new_height, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        x = (scaled_pix.width() - target_width) // 2
        y = (scaled_pix.height() - target_height) // 2
        return scaled_pix.copy(x, y, target_width, target_height)

    def get_circular_pixmap(self, pixmap, target_width: int, target_height: int):
        """
        Create a circular version of the given pixmap using an antialiased clip path.

        :param pixmap: Source pixmap already scaled/cropped to target size
        :param target_width: Target width
        :param target_height: Target height
        :return: Circular masked pixmap with transparent corners
        """
        from PySide6.QtGui import QPainter, QPainterPath, QPixmap  # local import to avoid altering global imports
        result = QPixmap(target_width, target_height)
        result.fill(Qt.transparent)

        painter = QPainter(result)
        painter.setRenderHint(QPainter.Antialiasing, True)
        path = QPainterPath()
        path.addEllipse(0, 0, target_width, target_height)
        painter.setClipPath(path)
        painter.drawPixmap(0, 0, pixmap)
        painter.end()

        return result

    def remove_avatar(self):
        """Remove the current avatar image."""
        self.avatar_preview.clear()
        self.remove_button.setEnabled(False)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.26 23:00:00                  #
# ================================================== #

OPENAI_COMPATIBLE_PROVIDERS = [
    "anthropic",
    "openai",
    "azure_openai",
    "google",
    "huggingface_router",
    "local_ai",
    "mistral_ai",
    "ollama",
    "perplexity",
    "deepseek_api",
    "x_ai",
    "open_router",
]

OPENAI_DISABLE_TOOLS = [
    "o1-mini",
    "o1-preview"
    "o4-mini-deep-research",
    "o3-deep-research",
]
OPENAI_REMOTE_TOOL_DISABLE_IMAGE = [
    "gpt-3.5-turbo",
    "o4-mini",
    "o4-mini-deep-research",
    "o3-deep-research",
    "codex-mini-latest",
]
OPENAI_REMOTE_TOOL_DISABLE_CODE_INTERPRETER = [
    "gpt-3.5-turbo",
    "codex-mini-latest",
]
OPENAI_REMOTE_TOOL_DISABLE_WEB_SEARCH = [
    "gpt-3.5-turbo",
    "codex-mini-latest",
]
OPENAI_REMOTE_TOOL_DISABLE_COMPUTER_USE = [
    "gpt-3.5-turbo",
    "o4-mini",
    "o4-mini-deep-research",
    "o3-deep-research",
    "codex-mini-latest",
]
OPENAI_REMOTE_TOOL_DISABLE_FILE_SEARCH = [
    "gpt-3.5-turbo",
    "codex-mini-latest",
]
OPENAI_REMOTE_TOOL_DISABLE_MCP = [
    "gpt-3.5-turbo",
    "codex-mini-latest",
]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.30 00:00:00                  #
# ================================================== #

import json

from pygpt_net.item.model import ModelItem


class Tools:
    def __init__(self, window=None):
        """
        Chat wrapper

        :param window: Window instance
        """
        self.window = window

    def prepare(
            self,
            model: ModelItem,
            functions: list
    ) -> list:
        """
        Prepare tools for the model (ChatCompletions API)

        :param model: Model item
        :param functions: List of functions
        :return: List of tools
        """
        tools = []
        if functions is not None and isinstance(functions, list):
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = {}
                if function['params'] is not None and function['params'] != "":
                    params = json.loads(function['params'])  # unpack JSON from string

                    # Google API fix:
                    if model.provider == "google":
                        if "properties" in params:
                            for k, v in params["properties"].items():
                                if "type" in v and v["type"] != "string":
                                    if "enum" in v:
                                        del v["enum"]

                tools.append({
                    "type": "function",
                    "function": {
                        "name": function['name'],
                        "parameters": params,
                        "description": function['desc'],
                    }
                })
        return tools

    def prepare_responses_api(
            self,
            model: ModelItem,
            functions: list
    ) -> list:
        """
        Prepare tools for the model (Responses API)

        :param model: Model item
        :param functions: List of functions
        :return: List of tools
        """
        tools = []
        if functions is not None and isinstance(functions, list):
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = {}
                if function['params'] is not None and function['params'] != "":
                    params = json.loads(function['params'])  # unpack JSON from string
                tools.append({
                    "type": "function",
                    "name": function['name'],
                    "parameters": params,
                    "description": function['desc'],
                })
        return tools


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.22 18:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, Slot, QUrl, QObject, Signal, QSize
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QVBoxLayout,
    QCheckBox,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QWidget,
    QSizePolicy,
)

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.textarea.editor import BaseCodeEditor
from pygpt_net.ui.widget.textarea.html import HtmlOutput
from pygpt_net.utils import trans


class ToolWidget:
    def __init__(self, window=None, tool=None):
        """
        HTML/JS canvas widget

        :param window: Window instance
        :param tool: Tool instance
        """
        self.window = window
        self.tool = tool  # tool instance
        self.output = None  # canvas output
        self.edit = None  # canvas edit
        self.btn_edit = None  # edit checkbox

        # --- Navigation bar state ---
        # This toolbar is shown only when opening via URL (open_url),
        # and hidden when using set_output/load_output.
        self.nav_bar = None
        self.nav_layout = None
        self.address_bar = None
        self.btn_back = None
        self.btn_next = None
        self.btn_reload = None
        self.btn_go = None

    def on_open(self):
        """On open"""
        pass

    def on_delete(self):
        """On delete"""
        if self.tool:
            self.tool.signals.update.disconnect(self.set_output)
            self.tool.signals.reload.disconnect(self.load_output)
            self.tool.signals.url.disconnect(self.open_url)  # keep connections clean
        if self.output:
            self.output.on_delete()

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.output.set_tab(tab)
        self.edit.set_tab(tab)

    def setup(self, all: bool = True) -> QVBoxLayout:
        """
        Setup widget body

        :param all: If True, setup all widgets
        :return: QVBoxLayout
        """
        self.output = CanvasOutput(self.window)
        self.edit = CanvasEdit(self.window)
        self.edit.setVisible(False)
        self.edit.textChanged.connect(
            lambda: self.tool.save_output()
        )

        # ---- Navigation bar ----
        # Visible only when navigating URLs via open_url or address bar.
        self.nav_bar = QWidget()
        self.nav_layout = QHBoxLayout(self.nav_bar)
        self.nav_layout.setContentsMargins(6, 4, 6, 4)  # small margins for compact look
        self.nav_layout.setSpacing(6)
        self.nav_bar.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)

        # Icon size and fixed navbar height (to prevent taking too much space)
        icon_size_px = 20  # base icon size for QPushButtons
        nav_height = max(32, min(44, icon_size_px + 16))  # compact, never half-screen
        self.nav_bar.setFixedHeight(nav_height)

        # Buttons
        self.btn_back = QPushButton()
        self.btn_back.setToolTip("Back")
        self.btn_back.setIcon(QIcon(":/icons/back.svg"))
        self.btn_back.setIconSize(QSize(icon_size_px, icon_size_px))
        self.btn_back.setFixedHeight(nav_height - 8)
        self.btn_back.setEnabled(False)
        self.btn_back.setAutoDefault(False)  # prevent Enter from triggering this
        try:
            self.btn_back.setDefault(False)
        except Exception:
            pass

        self.btn_next = QPushButton()
        self.btn_next.setToolTip("Next")
        self.btn_next.setIcon(QIcon(":/icons/forward.svg"))
        self.btn_next.setIconSize(QSize(icon_size_px, icon_size_px))
        self.btn_next.setFixedHeight(nav_height - 8)
        self.btn_next.setEnabled(False)
        self.btn_next.setAutoDefault(False)
        try:
            self.btn_next.setDefault(False)
        except Exception:
            pass

        self.btn_reload = QPushButton()
        self.btn_reload.setToolTip("Reload")
        self.btn_reload.setIcon(QIcon(":/icons/reload.svg"))
        self.btn_reload.setIconSize(QSize(icon_size_px, icon_size_px))
        self.btn_reload.setFixedHeight(nav_height - 8)
        self.btn_reload.setAutoDefault(False)
        try:
            self.btn_reload.setDefault(False)
        except Exception:
            pass

        # "Go" button
        self.btn_go = QPushButton()
        self.btn_go.setToolTip("Open URL")
        self.btn_go.setIcon(QIcon(":/icons/redo.svg"))
        self.btn_go.setIconSize(QSize(icon_size_px, icon_size_px))
        self.btn_go.setFixedHeight(nav_height - 8)
        self.btn_go.setAutoDefault(False)  # avoid stealing Enter
        try:
            self.btn_go.setDefault(False)
        except Exception:
            pass

        # Address bar (custom line edit to capture Enter and avoid default button reload)
        self.address_bar = AddressLineEdit(on_return_callback=self._on_address_enter)
        self.address_bar.setPlaceholderText("Enter URL and press Enter")
        self.address_bar.setFixedHeight(nav_height - 8)
        # Still keep the Qt signal for completeness (won't fire because we intercept the key)
        self.address_bar.returnPressed.connect(self._on_address_enter)

        # Hook up button actions
        self.btn_back.clicked.connect(lambda: self._navigate('back'))
        self.btn_next.clicked.connect(lambda: self._navigate('forward'))
        self.btn_reload.clicked.connect(lambda: self._navigate('reload'))
        self.btn_go.clicked.connect(self._on_address_enter)  # "Go" triggers address open

        # Build nav layout
        self.nav_layout.addWidget(self.btn_back)
        self.nav_layout.addWidget(self.btn_next)
        self.nav_layout.addWidget(self.btn_reload)
        self.nav_layout.addWidget(self.address_bar, 1)
        self.nav_layout.addWidget(self.btn_go)
        self.nav_bar.setVisible(False)  # hidden by default

        # edit checkbox
        self.btn_edit = QCheckBox(trans("html_canvas.btn.edit"))
        self.btn_edit.stateChanged.connect(
            lambda: self.tool.toggle_edit(self)
        )

        path = self.tool.get_current_path()
        path_label = HelpLabel(path)
        path_label.setMaximumHeight(30)
        path_label.setAlignment(Qt.AlignRight)

        bottom_layout = QHBoxLayout()
        bottom_layout.addWidget(self.btn_edit)
        bottom_layout.addWidget(path_label)

        output_layout = QVBoxLayout()
        # put navigation bar above the web output
        output_layout.addWidget(self.nav_bar)
        output_layout.addWidget(self.output)
        output_layout.addWidget(self.edit)
        output_layout.setContentsMargins(0, 0, 0, 0)

        # connect signals
        self.output.signals.save_as.connect(
            self.tool.handle_save_as)
        self.output.signals.audio_read.connect(
            self.window.controller.chat.render.handle_audio_read)

        # Keep nav bar in sync with the underlying view if signals are available
        # (urlChanged/loadStarted/loadFinished are part of QWebEngineView)
        if hasattr(self.output, 'urlChanged'):
            self.output.urlChanged.connect(self._on_url_changed)
        if hasattr(self.output, 'loadStarted'):
            try:
                self.output.loadStarted.connect(lambda: self._update_nav_controls())
            except TypeError:
                pass
        if hasattr(self.output, 'loadFinished'):
            try:
                self.output.loadFinished.connect(lambda ok: self._update_nav_controls())
            except TypeError:
                pass

        self.tool.signals.update.connect(self.set_output)
        self.tool.signals.reload.connect(self.load_output)
        self.tool.signals.url.connect(self.open_url)

        layout = QVBoxLayout()
        layout.addLayout(output_layout)
        layout.addLayout(bottom_layout)
        return layout

    @Slot(str)
    def set_output(self, content: str):
        """
        Set output content

        :param content: Content
        """
        self.edit.setPlainText(content)
        # Hide navigation bar when content is set via set_output (not a URL)
        self._show_navbar(False)

    @Slot(str)
    def open_url(self, url: str):
        """
        Open URL in output

        :param url: URL
        """
        print("Opening URL:", url)
        self._show_navbar(True)  # Show navigation bar for URL navigation
        if self.address_bar:
            self.address_bar.setText(url)
        self.output.setUrl(QUrl(url))
        self._update_nav_controls()

    @Slot(str)
    def load_output(self, path: str):
        """
        Load output content

        :param path: Content
        """
        self.output.setUrl(QUrl().fromLocalFile(path))
        # Hide navigation bar when loading from local file/path
        self._show_navbar(False)

    # ------------------
    # Navigation helpers
    # ------------------
    def _show_navbar(self, show: bool):
        """Show/hide the navigation bar."""
        if self.nav_bar:
            self.nav_bar.setVisible(show)
        # Keep controls state up-to-date
        self._update_nav_controls()

    def _navigate(self, action: str):
        """Perform navigation action on the output view."""
        if not self.output:
            return
        try:
            if action == 'back' and hasattr(self.output, 'back'):
                self.output.back()
            elif action == 'forward' and hasattr(self.output, 'forward'):
                self.output.forward()
            elif action == 'reload' and hasattr(self.output, 'reload'):
                self.output.reload()
        finally:
            self._update_nav_controls()

    def _on_address_enter(self):
        """Handle Enter pressed in the address bar or clicking GO."""
        if not self.output or not self.address_bar:
            return
        text = self.address_bar.text().strip()
        if not text:
            return
        url = QUrl.fromUserInput(text)
        if url.isValid():
            self._show_navbar(True)
            self.output.setUrl(url)
            self._update_nav_controls()

    def _on_url_changed(self, url: QUrl):
        """Keep address bar in sync with the view URL."""
        if self.address_bar:
            self.address_bar.setText(url.toString())
        self._update_nav_controls()

    def _update_nav_controls(self):
        """Enable/disable back/forward buttons based on history."""
        if not self.output:
            return
        # Reload is generally available when navbar is visible
        if self.btn_reload:
            self.btn_reload.setEnabled(True)
        # Back/forward depend on history if available
        if self.btn_back and self.btn_next:
            try:
                hist = self.output.history()
                self.btn_back.setEnabled(bool(hist and hist.canGoBack()))
                self.btn_next.setEnabled(bool(hist and hist.canGoForward()))
            except Exception:
                # If history is not available, keep safe defaults
                self.btn_back.setEnabled(False)
                self.btn_next.setEnabled(False)


class CanvasOutput(HtmlOutput):
    def __init__(self, window=None):
        """
        HTML canvas output

        :param window: main window
        """
        super(CanvasOutput, self).__init__(window)
        self.window = window


class CanvasEdit(BaseCodeEditor):
    def __init__(self, window=None):
        """
        Python interpreter output

        :param window: main window
        """
        super(CanvasEdit, self).__init__(window)
        self.window = window
        self.setReadOnly(False)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'interpreter-output')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)


# --- Address bar input widget ---
class AddressLineEdit(QLineEdit):
    """
    Custom QLineEdit that ensures Enter triggers opening the typed URL
    and prevents default buttons (e.g., Reload) from consuming the key.
    """
    def __init__(self, on_return_callback=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_return_callback = on_return_callback

    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Return, Qt.Key_Enter):
            # Call the assigned handler and stop propagation so default buttons won't trigger.
            if callable(self._on_return_callback):
                self._on_return_callback()
            event.accept()
            return
        super().keyPressEvent(event)


class ToolSignals(QObject):
    update = Signal(str)  # data
    reload = Signal(str)  # path
    url = Signal(str)  # url


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 17:00:00                  #
# ================================================== #

from typing import Any

from pygpt_net.utils import trans


class Batch:
    def __init__(self, window=None):
        """
        Assistants batch controller

        :param window: Window instance
        """
        self.window = window
        self.files_to_upload = []

    def import_assistants(self, force: bool = False):
        """
        Import all remote assistants from API

        :param force: if true, imports without confirmation
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='assistant.import',
                id='',
                msg=trans('confirm.assistant.import'),
            )
            return

        # run asynchronous
        self.window.update_status("Importing assistants...please wait...")
        self.window.core.api.openai.assistants.importer.import_assistants()

    def handle_imported_assistants(self, num: int):
        """
        Handle imported assistants

        :param num: number of imported assistants
        """
        self.window.controller.assistant.update()
        self.window.controller.remote_store.update()
        self.window.controller.assistant.files.update()
        self.window.update_status("OK. Imported assistants: " + str(num) + ".")
        self.window.ui.dialogs.alert(trans("status.finished"))

    def handle_imported_assistants_failed(self, error: Any):
        """
        Handle error on importing assistants

        :param error: error message
        """
        self.window.core.debug.log(error)
        print("Error importing assistants")
        self.window.controller.assistant.update()
        self.window.update_status("Error importing assistants.")
        self.window.ui.dialogs.alert(error)

    def remove_store_from_assistants(self, store_id: str):
        """
        Remove vector store from all assistants after store deletion

        :param store_id: vector store ID
        """
        for id in list(self.window.core.assistants.items.keys()):
            assistant = self.window.core.assistants.get_by_id(id)
            if assistant is not None:
                if assistant.vector_store == store_id:
                    assistant.vector_store = None  # remove from assistant

        self.window.core.assistants.save()
        self.window.core.remote_store.openai.files.on_store_deleted(store_id)  # remove from files

    def remove_all_stores_from_assistants(self):
        """Remove all vector stores from all assistants"""
        for id in list(self.window.core.assistants.items.keys()):
            assistant = self.window.core.assistants.get_by_id(id)
            if assistant is not None:
                assistant.vector_store = None

        self.window.core.assistants.save()
        self.window.core.remote_store.openai.files.on_all_stores_deleted()  # remove all from files

    def handle_status_change(self, mode: str, msg: str):
        """
        Handle status change

        :param mode: status mode
        :param msg: status message
        """
        self.window.update_status(msg)